import os

from indy_common.config import GENERAL_CONFIG_DIR


def create_config_dirs(base_dir):
    gen_conf_dir_default = GENERAL_CONFIG_DIR.lstrip('/') \
        if GENERAL_CONFIG_DIR.startswith('/') else GENERAL_CONFIG_DIR

    general_config_dir = os.path.join(base_dir, gen_conf_dir_default)
    os.makedirs(general_config_dir)

    return general_config_dir



import json

import pytest
from indy.anoncreds import issuer_create_and_store_credential_def, issuer_create_schema
from indy.ledger import build_cred_def_request, build_schema_request

from indy_common.constants import REF
from indy_node.test.helper import modify_field
from plenum.common.exceptions import RequestRejectedException
from plenum.common.types import OPERATION
from plenum.test.helper import sdk_sign_and_submit_req, sdk_get_and_check_replies, max_3pc_batch_limits


def sdk_send_claim_def(looper, sdk_pool_handle, sdk_wallet, tag, schema_json):
    wallet_handle, identifier = sdk_wallet

    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, schema_json, tag, "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    reply = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet, request)])
    return reply


def test_send_claim_def_succeeds(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, schema_json):
    wallet_handle, identifier = sdk_wallet_trustee

    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, schema_json, "some_tag", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])


def test_send_claim_def_succeeds_for_large_schema(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, large_schema_json):
    wallet_handle, identifier = sdk_wallet_trustee

    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, large_schema_json, "some_tag", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])


def test_send_claim_def_schema_and_claim_def_in_one_batch(
        looper, tconf, sdk_pool_handle, nodeSet, sdk_wallet_trustee):
    with max_3pc_batch_limits(tconf, size=2) as tconf:
        initial_seq_no = nodeSet[0].domainLedger.size
        wallet_handle, identifier = sdk_wallet_trustee

        # send SCHEMA
        _, schema_json = looper.loop.run_until_complete(
            issuer_create_schema(identifier, "name2", "2.0", json.dumps(["first", "last"])))
        schema_req_json = looper.loop.run_until_complete(
            build_schema_request(identifier, schema_json))
        schema_req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, schema_req_json)

        # send CLAIM_DEF
        _, claim_def_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
            wallet_handle, identifier, schema_json, "some_tag", "CL", json.dumps({"support_revocation": True})))
        claim_def_req_json = looper.loop.run_until_complete(
            build_cred_def_request(identifier, claim_def_json))
        claim_def_req_json = modify_field(claim_def_req_json, initial_seq_no + 1, OPERATION, REF)
        claim_def_req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, claim_def_req_json)

        # check both are written
        sdk_get_and_check_replies(looper, [schema_req, claim_def_req])


def test_send_claim_def_fails_if_ref_is_seqno_of_non_schema_txn(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, schema_json):
    wallet_handle, identifier = sdk_wallet_trustee

    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, schema_json, "some_tag1", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    request = modify_field(request, 1, OPERATION, REF)
    with pytest.raises(RequestRejectedException) as e:
        sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    e.match('isn\'t seqNo of the schema.')


def test_send_claim_def_fails_if_ref_is_not_existing_seqno(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, schema_json):
    wallet_handle, identifier = sdk_wallet_trustee

    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, schema_json, "some_tag2", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    request = modify_field(request, 999999, OPERATION, REF)
    with pytest.raises(RequestRejectedException) as e:
        sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    e.match('doesn\'t exist')


def test_update_claim_def_for_same_schema_and_signature_type(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, schema_json):
    wallet_handle, identifier = sdk_wallet_trustee

    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, schema_json, "some_tag3", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])

    definition_json = modify_field(definition_json, '999', 'value', 'primary', 'n')
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])


def test_can_send_same_claim_def_by_different_issuers(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, sdk_wallet_steward, schema_json):
    wallet_handle, identifier = sdk_wallet_trustee

    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, schema_json, "some_tag4", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])

    wallet_handle, identifier = sdk_wallet_steward
    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, schema_json, "some_tag4", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, request)])



import multiprocessing
import os

import pytest

from stp_core.loop.eventually import eventually
from indy_common.constants import APP_NAME
from indy_common.version import src_version_cls
from indy_node.utils.migration_tool import _get_current_platform
from indy_node.utils.node_control_utils import NodeControlUtil, DebianVersion

from indy_node.test.upgrade.helper import (
    NodeControlToolExecutor as NCT,
    sendUpgradeMessage,
    nodeControlGeneralMonkeypatching,
    releaseVersion,
    bumpedVersion
)

m = multiprocessing.Manager()
# TODO why do we expect that
whitelist = ['Unexpected error in _upgrade test']


@pytest.fixture(scope='module')
def tconf(tconf, tdir):
    tconf.LOG_DIR = tdir
    yield tconf


def testNodeControlPerformsMigrations(monkeypatch, tdir, looper, tconf):
    version = bumpedVersion()
    stdout = 'teststdout'
    migrationFile = 'migrationProof'
    migrationText = "{} {}".format(releaseVersion(), version)

    old_call_upgrade_script = None

    def mock_call_upgrade_script(*args, **kwargs):
        old_call_upgrade_script(*args, **kwargs)
        monkeypatch.setattr(
            NodeControlUtil,
            '_get_curr_info',
            lambda *x, **y: (
                "Package: {}\nVersion: {}\n".format(APP_NAME, version)
            )
        )

    def mock_migrate(curr_src_ver: str, new_src_ver: str):
        with open(os.path.join(tdir, migrationFile), 'w') as f:
            f.write("{} {}".format(curr_src_ver, new_src_ver))

    def transform(tool):
        nonlocal old_call_upgrade_script
        nodeControlGeneralMonkeypatching(tool, monkeypatch, tdir, stdout)
        monkeypatch.setattr(tool, '_do_migration', mock_migrate)
        old_call_upgrade_script = getattr(tool, '_call_upgrade_script')
        monkeypatch.setattr(tool, '_call_upgrade_script', mock_call_upgrade_script)

    def checkMigration():
        with open(os.path.join(tdir, migrationFile)) as f:
            assert f.read() == migrationText

    monkeypatch.setattr(
        NodeControlUtil, 'get_latest_pkg_version',
        lambda *x, **y: DebianVersion(
            version, upstream_cls=src_version_cls())
    )

    nct = NCT(backup_dir=tdir, backup_target=tdir, transform=transform, config=tconf)
    try:
        sendUpgradeMessage(version)
        looper.run(eventually(checkMigration))
    finally:
        nct.stop()


def test_get_current_platform():
    _get_current_platform()



import pytest

from indy_common.constants import RS_ID, GET_RICH_SCHEMA_OBJECT_BY_ID, RS_VERSION, RS_NAME, RS_CONTENT, RS_TYPE
from indy_common.types import Request
from indy_node.server.request_handlers.read_req_handlers.rich_schema.get_rich_schema_object_by_id_handler import \
    GetRichSchemaObjectByIdHandler
from indy_node.test.state_proof.helper import check_valid_proof
from indy_node.test.state_proof.test_state_multi_proofs_for_get_requests import is_proof_verified, save_multi_sig
from plenum.common.constants import TXN_TYPE, OP_VER, DATA, CURRENT_PROTOCOL_VERSION
from plenum.common.txn_util import reqToTxn, append_txn_metadata, get_payload_data
from plenum.common.util import randomString, SortedDict


@pytest.fixture()
def id():
    return randomString()


@pytest.fixture()
def get_rich_schema_req(id):
    return Request(identifier=randomString(),
                   reqId=1234,
                   sig="sig",
                   protocolVersion=CURRENT_PROTOCOL_VERSION,
                   operation={
                       TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_ID,
                       RS_ID: id,
                   })


@pytest.fixture()
def get_rich_schema_by_id_handler(db_manager):
    return GetRichSchemaObjectByIdHandler(db_manager)


def test_get_rich_schema_obj(db_manager, handler_and_request, id,
                             get_rich_schema_by_id_handler, get_rich_schema_req):
    # prepare: store object in state with bls multi-sg
    handler, request = handler_and_request
    op = request.operation
    op[RS_ID] = id
    seq_no = 1
    txn_time = 1560241033
    txn = reqToTxn(request)
    append_txn_metadata(txn, seq_no, txn_time)
    handler.update_state(txn, None, request)
    handler.state.commit()
    save_multi_sig(db_manager)

    # execute: get object
    result = get_rich_schema_by_id_handler.get_result(get_rich_schema_req)

    # check
    assert result
    expected_data = SortedDict({
        'ver': op[OP_VER],
        'id': op[RS_ID],
        'rsType': op[RS_TYPE],
        'rsName': op[RS_NAME],
        'rsVersion': op[RS_VERSION],
        'content': op[RS_CONTENT],
        'from': request.identifier,
        'endorser': request.endorser,
    })
    assert SortedDict(result['data']) == expected_data
    assert result['seqNo'] == seq_no
    assert result['txnTime'] == txn_time
    assert result['state_proof']
    check_valid_proof(result)
    path = id.encode()
    assert is_proof_verified(db_manager,
                             result['state_proof'],
                             path, result[DATA], seq_no, txn_time)


def test_get_rich_schema_obj_not_existent(db_manager, handler_and_request,
                                          get_rich_schema_by_id_handler, get_rich_schema_req):
    save_multi_sig(db_manager)

    # execute: get object
    result = get_rich_schema_by_id_handler.get_result(get_rich_schema_req)

    # check
    assert result
    assert result['data'] is None
    assert result['seqNo'] is None
    assert result['txnTime'] is None
    assert result['state_proof']
    check_valid_proof(result)


def test_get_rich_schema_obj_committed_only(db_manager, handler_and_request, id,
                                            get_rich_schema_by_id_handler, get_rich_schema_req):
    # prepare: store object in state with bls multi-sig, and then update the object (uncommitted)
    handler, request = handler_and_request
    op = request.operation
    op[RS_ID] = id
    seq_no = 1
    txn_time = 1560241033
    txn = reqToTxn(request)
    append_txn_metadata(txn, seq_no, txn_time)
    handler.update_state(txn, None, request)
    handler.state.commit()

    get_payload_data(txn)[RS_NAME] = "new uncommitted name"
    get_payload_data(txn)[RS_VERSION] = "3.0"
    handler.update_state(txn, None, request)

    save_multi_sig(db_manager)

    # execute: get object
    result = get_rich_schema_by_id_handler.get_result(get_rich_schema_req)

    # check
    assert result
    expected_data = SortedDict({
        'ver': op[OP_VER],
        'id': op[RS_ID],
        'rsType': op[RS_TYPE],
        'rsName': op[RS_NAME],
        'rsVersion': op[RS_VERSION],
        'content': op[RS_CONTENT],
        'from': request.identifier,
        'endorser': request.endorser,
    })
    assert SortedDict(result['data']) == expected_data
    assert result['seqNo'] == seq_no
    assert result['txnTime'] == txn_time
    assert result['state_proof']
    check_valid_proof(result)
    path = id.encode()
    assert is_proof_verified(db_manager,
                             result['state_proof'],
                             path, result[DATA], seq_no, txn_time)



import json

import pytest


@pytest.fixture(scope="function", params=[False, True])
def with_verkey(request):
    return request.param


@pytest.fixture
def diddoc_content():
    yield {
        "@context": [
            "https://www.w3.org/ns/did/v1",
            "https://identity.foundation/didcomm-messaging/service-endpoint/v1",
        ],
        "serviceEndpoint": [
            {
                "id": "did:indy:sovrin:123456#didcomm",
                "type": "didcomm-messaging",
                "serviceEndpoint": "https://example.com",
                "recipientKeys": ["#verkey"],
                "routingKeys": [],
            }
        ],
    }


@pytest.fixture
def diddoc_content_json(diddoc_content):
    yield json.dumps(diddoc_content)



import pytest
import time

from indy_common.authorize.auth_constraints import AuthConstraint, \
    AuthConstraintAnd, \
    AuthConstraintOr, ConstraintsEnum
from indy_common.authorize.authorizer import CompositeAuthorizer, RolesAuthorizer, AndAuthorizer, OrAuthorizer, \
    AuthValidationError
from indy_common.types import Request
from indy_node.persistence.idr_cache import IdrCache
from plenum.common.constants import CURRENT_PROTOCOL_VERSION, STEWARD, TRUSTEE
from plenum.test.helper import randomOperation
from storage.kv_in_memory import KeyValueStorageInMemory


@pytest.fixture(scope='module')
def req_auth():
    return Request(identifier="some_identifier",
                   reqId=1,
                   operation=randomOperation(),
                   signature="signature",
                   protocolVersion=CURRENT_PROTOCOL_VERSION)


@pytest.fixture(scope='module')
def idr_cache(req_auth):
    cache = IdrCache("Cache",
                     KeyValueStorageInMemory())
    cache.set(req_auth.identifier, 1, int(time.time()), role=STEWARD,
              verkey="SomeVerkey", isCommitted=False)
    return cache


@pytest.fixture(scope='module')
def composite_authorizer(idr_cache):
    authorizer = CompositeAuthorizer()
    authorizer.register_authorizer(RolesAuthorizer(idr_cache))
    authorizer.register_authorizer(AndAuthorizer(), ConstraintsEnum.AND_CONSTRAINT_ID)
    authorizer.register_authorizer(OrAuthorizer(), ConstraintsEnum.OR_CONSTRAINT_ID)
    return authorizer


@pytest.fixture(scope='module')
def other_req_id():
    return Request(identifier="some_other_identifier",
                   reqId=2,
                   operation=randomOperation(),
                   signature="signature",
                   protocolVersion=CURRENT_PROTOCOL_VERSION)


def test_authorize_with_role(composite_authorizer, req_auth):
    assert composite_authorizer.authorize(req_auth, AuthConstraint(STEWARD, 1), None)


def test_raise_on_not_authorize_with_other_role(composite_authorizer, req_auth):
    with pytest.raises(AuthValidationError):
        assert composite_authorizer.authorize(req_auth, AuthConstraint(TRUSTEE, 1), None)


def test_authorize_with_and_constraint(composite_authorizer, req_auth):
    composite_authorizer.authorize(req_auth,
                                   AuthConstraintAnd([AuthConstraint(STEWARD, 1), AuthConstraint(STEWARD, 1)]),
                                   None)


def test_not_authorize_with_and_constraint(composite_authorizer, req_auth):
    with pytest.raises(AuthValidationError):
        composite_authorizer.authorize(req_auth,
                                       AuthConstraintAnd([AuthConstraint(STEWARD, 1), AuthConstraint(TRUSTEE, 1)]),
                                       None)


def test_authorize_with_or_constraint(composite_authorizer, req_auth):
    composite_authorizer.authorize(req_auth,
                                   AuthConstraintOr([AuthConstraint(STEWARD, 1), AuthConstraint(STEWARD, 1)]),
                                   None)


def test_authorize_with_or_constraint_with_one_fail(composite_authorizer, req_auth):
    composite_authorizer.authorize(req_auth,
                                   AuthConstraintOr([AuthConstraint(STEWARD, 1), AuthConstraint(TRUSTEE, 1)]),
                                   None)


def test_not_authorized_with_or_constraint(composite_authorizer, req_auth):
    with pytest.raises(AuthValidationError):
        composite_authorizer.authorize(req_auth,
                                       AuthConstraintOr([AuthConstraint(TRUSTEE, 1), AuthConstraint(TRUSTEE, 1)]),
                                       None)



from copy import deepcopy

from indy_common.constants import NODE

from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from plenum.common.constants import DATA, SERVICES, VALIDATOR, TARGET_NYM, ALIAS, BLS_KEY_PROOF
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.node_handler import NodeHandler as PNodeHandler


class NodeHandler(PNodeHandler):

    def __init__(self, database_manager: DatabaseManager, bls_crypto_verifier,
                 write_req_validator: WriteRequestValidator):
        super().__init__(database_manager, bls_crypto_verifier)
        self.write_req_validator = write_req_validator

    def _is_steward(self, nym, is_committed: bool = True):
        return self.database_manager.idr_cache.hasSteward(nym, is_committed)

    def _get_node_data(self, nym, is_committed: bool = True):
        key = nym.encode()
        data = self.state.get(key, is_committed)
        if not data:
            return {}
        return self.state_serializer.deserialize(data)

    def _auth_error_while_adding_node(self, request):
        origin = request.identifier
        operation = request.operation
        data = operation.get(DATA, {})
        error = self._data_error_while_validating(data, skip_keys=False)
        if error:
            return error

        if self._steward_has_node(origin):
            return "{} already has a node".format(origin)
        error = self._is_node_data_conflicting(data)
        if error:
            return "existing data has conflicts with " \
                   "request data {}. Error: {}".format(operation.get(DATA), error)
        self.write_req_validator.validate(request,
                                          [AuthActionAdd(txn_type=NODE,
                                                         field=SERVICES,
                                                         value=data.get(SERVICES, [VALIDATOR]))])

    def _auth_error_while_updating_node(self, request):
        origin = request.identifier
        is_trustee = self.database_manager.idr_cache.hasTrustee(origin, isCommitted=False)
        if not is_trustee:
            error = super()._auth_error_while_updating_node(request)
            if error:
                return error
        origin = request.identifier
        operation = request.operation
        node_nym = operation.get(TARGET_NYM)

        data = operation.get(DATA, {})
        error = self._data_error_while_validating_update(data, node_nym)
        if error:
            return error

        is_steward_of_node = self._is_steward_of_node(
            origin, node_nym, is_committed=False)

        node_info = self.get_from_state(node_nym, is_committed=False)
        data = deepcopy(data)
        data.pop(ALIAS, None)
        for k in data:
            if k == BLS_KEY_PROOF:
                continue
            old_val = node_info.get(k, None) if node_info else None
            new_val = data[k]
            if k == SERVICES:
                if not old_val:
                    old_val = []
                if not new_val:
                    new_val = []
            if old_val != new_val:
                self.write_req_validator.validate(request,
                                                  [AuthActionEdit(txn_type=NODE,
                                                                  field=k,
                                                                  old_value=old_val,
                                                                  new_value=new_val,
                                                                  is_owner=is_steward_of_node)])



import json
from collections import namedtuple
from typing import Iterable

import pytest
from indy.ledger import build_pool_config_request, build_nym_request

from plenum.common.messages.node_messages import PrePrepare, Prepare, Commit
from plenum.common.request import Request
from plenum.common.signer_simple import SimpleSigner
from plenum.common.util import compare_3PC_keys
from plenum.server.catchup.node_leecher_service import NodeLeecherService
from plenum.test.delayers import DEFAULT_DELAY, icDelay, cr_delay
from plenum.test.helper import max_3pc_batch_limits, sdk_get_replies, sdk_check_reply, \
    random_string, sdk_gen_pool_request, sdk_sign_and_submit_req_obj
from plenum.test.node_catchup.helper import ensure_all_nodes_have_same_data
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.stasher import start_delaying, stop_delaying_and_process, delay_rules
from plenum.test.test_node import ensureElectionsDone
from plenum.test.view_change_service.helper import trigger_view_change
from stp_core.loop.eventually import eventually


@pytest.fixture(scope="module")
def tconf(tconf):
    with max_3pc_batch_limits(tconf, size=1) as tconf:
        yield tconf


@pytest.fixture(scope='module')
def sdk_wallet_new_steward(looper, sdk_pool_handle, sdk_wallet_trustee):
    wh, client_did = sdk_add_new_nym(looper, sdk_pool_handle,
                                     sdk_wallet_trustee,
                                     alias='new_steward_qwerty',
                                     role='STEWARD')
    return wh, client_did


def test_view_change_during_alternating_unstash(looper, txnPoolNodeSet, sdk_pool_handle,
                                                sdk_wallet_trustee, sdk_wallet_new_steward, tconf):
    slow_node = txnPoolNodeSet[-1]
    other_nodes = txnPoolNodeSet[:-1]

    slow_stasher = slow_node.nodeIbStasher
    other_stashers = [n.nodeIbStasher for n in other_nodes]
    all_stashers = [n.nodeIbStasher for n in txnPoolNodeSet]

    # Ensure that pool has expected 3PC state
    for node in txnPoolNodeSet:
        assert node.master_replica.last_ordered_3pc == (0, 1)

    # Prevent ordering of some requests
    start_delaying(all_stashers, delay_3pc_after(0, 7, Prepare, Commit))

    # Stop ordering on slow node and send requests
    slow_node_after_5 = start_delaying(slow_stasher, delay_3pc_after(0, 5, Commit))
    slow_node_until_5 = start_delaying(slow_stasher, delay_3pc_after(0, 0))
    reqs_view_0 = sdk_send_alternating_ledgers_requests(looper, sdk_pool_handle,
                                                        sdk_wallet_trustee, sdk_wallet_new_steward, 8)

    # Make pool order first 2 batches and pause
    pool_after_3 = start_delaying(other_stashers, delay_3pc_after(0, 3))
    looper.run(eventually(check_nodes_ordered_till, other_nodes, 0, 3))

    # Start catchup, continue ordering everywhere (except two last batches on slow node)
    with delay_rules(slow_stasher, cr_delay()):
        slow_node._do_start_catchup(just_started=False)
        looper.run(eventually(check_catchup_is_started, slow_node))
        stop_delaying_and_process(pool_after_3)
        looper.run(eventually(check_nodes_ordered_till, other_nodes, 0, 7))

    # Finish catchup and continue processing on slow node
    looper.run(eventually(check_catchup_is_finished, slow_node))
    stop_delaying_and_process(slow_node_until_5)
    looper.run(eventually(check_nodes_ordered_till, [slow_node], 0, 5))

    # Start view change and allow slow node to get remaining commits
    with delay_rules(all_stashers, icDelay()):
        trigger_view_change(txnPoolNodeSet)
        looper.runFor(0.1)
    stop_delaying_and_process(slow_node_after_5)

    # Ensure that expected number of requests was ordered
    replies = sdk_get_replies(looper, reqs_view_0)
    for req in replies[:6]:
        sdk_check_reply(req)

    # Ensure that everything is ok
    ensureElectionsDone(looper, txnPoolNodeSet)
    ensure_all_nodes_have_same_data(looper, txnPoolNodeSet)


def sdk_gen_domain_request(looper, sdk_wallet):
    _, did = sdk_wallet
    target_did = SimpleSigner(seed=random_string(32).encode()).identifier
    req = looper.loop.run_until_complete(build_nym_request(did, target_did, None, None, None))
    return Request(**json.loads(req))


def sdk_gen_config_request(looper, sdk_wallet):
    _, did = sdk_wallet
    req = looper.loop.run_until_complete(build_pool_config_request(did, True, False))
    return Request(**json.loads(req))


def sdk_send_alternating_ledgers_requests(looper, pool_h, sdk_wallet_trustee, sdk_wallet_new_steward, count: int):
    node_alias = random_string(7)
    node_did = SimpleSigner(seed=random_string(32).encode()).identifier

    ReqFactory = namedtuple('ReqFactory', 'wallet create')
    req_factories = [
        ReqFactory(wallet=sdk_wallet_new_steward,
                   create=lambda: sdk_gen_pool_request(looper, sdk_wallet_new_steward, node_alias, node_did)),
        ReqFactory(wallet=sdk_wallet_trustee,
                   create=lambda: sdk_gen_domain_request(looper, sdk_wallet_trustee)),
        ReqFactory(wallet=sdk_wallet_trustee,
                   create=lambda: sdk_gen_config_request(looper, sdk_wallet_trustee))
    ]

    res = []
    for i in range(count):
        factory = req_factories[i % len(req_factories)]

        req = factory.create()
        res.append(sdk_sign_and_submit_req_obj(looper, pool_h, factory.wallet, req))
        looper.runFor(0.1)  # Give nodes some time to start ordering, so that requests are really alternating
    return res


def delay_3pc_after(view_no, pp_seq_no, *args):
    if not args:
        args = (PrePrepare, Prepare, Commit)

    def _delayer(msg_frm):
        msg, frm = msg_frm
        if not isinstance(msg, args):
            return
        if msg.viewNo != view_no:
            return
        if msg.ppSeqNo <= pp_seq_no:
            return
        return DEFAULT_DELAY

    _delayer.__name__ = "delay_3pc_after({}, {}, {})".format(view_no, pp_seq_no, args)
    return _delayer


def check_nodes_ordered_till(nodes: Iterable, view_no: int, pp_seq_no: int):
    for node in nodes:
        assert compare_3PC_keys((view_no, pp_seq_no), node.master_replica.last_ordered_3pc) >= 0


def check_catchup_is_started(node):
    assert node.ledgerManager._node_leecher._state != NodeLeecherService.State.Idle


def check_catchup_is_finished(node):
    assert node.ledgerManager._node_leecher._state == NodeLeecherService.State.Idle



import multiprocessing
import os
import functools
import shutil

import time

from stp_core.loop.eventually import eventually
from indy_common.version import src_version_cls
from indy_node.server.upgrader import Upgrader
from indy_node.utils.node_control_utils import NodeControlUtil, DebianVersion

from indy_node.test.upgrade.helper import (
    NodeControlToolExecutor as NCT,
    sendUpgradeMessage,
    nodeControlGeneralMonkeypatching,
    bumpedVersion
)

m = multiprocessing.Manager()
# TODO why do we expect that
whitelist = ['Unexpected error in _upgrade test']


def testNodeControlRestoresFromBackups(monkeypatch, tdir, looper, tconf):
    version = bumpedVersion()
    stdout = 'teststdout'
    backupWasRestored = m.Value('b', False)
    testFile = 'testFile'
    original_text = '1'
    new_text = '2'

    def testRestoreBackup(tool, src_ver: str):
        tool._restore_from_backup_test(src_ver)
        backupWasRestored.value = True

    def mockMigrate(tool, *args):
        monkeypatch.setattr(tool, '_do_migration', lambda *args: None)
        with open(os.path.join(tool.indy_dir, testFile), 'w') as f:
            f.write(new_text)
        raise Exception('test')

    def transform(tool):
        nodeControlGeneralMonkeypatching(tool, monkeypatch, tdir, stdout)
        tool._restore_from_backup_test = tool._restore_from_backup
        monkeypatch.setattr(
            tool, '_do_migration', functools.partial(mockMigrate, tool))
        monkeypatch.setattr(tool, '_restore_from_backup',
                            functools.partial(testRestoreBackup, tool))

    def checkBackupRestored(tool):
        assert backupWasRestored.value

    monkeypatch.setattr(
        NodeControlUtil, 'get_latest_pkg_version',
        lambda *x, **y: DebianVersion(
            version, upstream_cls=src_version_cls())
    )

    nct = NCT(backup_dir=tdir, backup_target=tdir, transform=transform)
    try:
        with open(os.path.join(nct.tool.indy_dir, testFile), 'w') as f:
            f.write(original_text)
        sendUpgradeMessage(version)
        looper.run(eventually(checkBackupRestored, nct.tool))
        with open(os.path.join(nct.tool.indy_dir, testFile)) as f:
            assert original_text == f.read()
    finally:
        clean_dir(nct.tool.base_dir)
        nct.stop()


def clean_dir(dir):
    for file in os.listdir(dir):
        file = os.path.join(dir, file)
        if os.path.isfile(file):
            os.remove(file)
        if os.path.isdir(file):
            shutil.rmtree(file, ignore_errors=True)



import multiprocessing

import pytest

from indy_node.test.upgrade.helper import NodeControlToolExecutor, nodeControlGeneralMonkeypatching
from stp_core.loop.eventually import eventually

RESTART_TIMEOUT = 10


@pytest.fixture(scope="module")
def tconf(tconf, tdir):
    old_restart_timeout = tconf.INCONSISTENCY_WATCHER_NETWORK_TIMEOUT
    tconf.INCONSISTENCY_WATCHER_NETWORK_TIMEOUT = RESTART_TIMEOUT
    tconf.LOG_DIR = tdir
    yield tconf
    tconf.INCONSISTENCY_WATCHER_NETWORK_TIMEOUT = old_restart_timeout


def test_restart_on_inconsistency(looper, txnPoolNodeSet, tconf, tdir, monkeypatch):
    restarted = multiprocessing.Value('i', 0)

    def notify_restart():
        with restarted.get_lock():
            restarted.value = 1

    def transform(tool):
        nodeControlGeneralMonkeypatching(tool, monkeypatch, tdir, 'teststdout')
        monkeypatch.setattr(tool, '_restart', notify_restart)

    def check_restart(value):
        with restarted.get_lock():
            assert restarted.value == value

    nct = NodeControlToolExecutor(backup_dir=tdir, backup_target=tdir, transform=transform, config=tconf)
    try:
        # Trigger inconsistent 3PC state event
        txnPoolNodeSet[0].on_inconsistent_3pc_state()

        # Ensure we don't restart before timeout
        looper.runFor(0.8 * RESTART_TIMEOUT)
        check_restart(0)

        # Ensure we restart eventually
        looper.run(eventually(check_restart, 1, retryWait=1.0))
    finally:
        nct.stop()





import asyncio
from datetime import datetime, timedelta

import dateutil
from jsonpickle import json

from indy_node.server.restart_log import RestartLog, RestartLogData

from indy_common.constants import POOL_RESTART, ACTION, START, CANCEL
from indy_node.server.restarter import Restarter
from indy_node.test.pool_restart.helper import _createServer, _stopServer, sdk_send_restart
from plenum.common.constants import REPLY, TXN_TYPE
from plenum.common.types import f


def test_pool_restart(
        sdk_pool_handle, sdk_wallet_trustee, looper, tconf, txnPoolNodeSet):

    server, indicator = looper.loop.run_until_complete(
        _createServer(
            host=tconf.controlServiceHost,
            port=tconf.controlServicePort
        )
    )

    unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    start_at = unow + timedelta(seconds=1000)
    req_obj, responses = sdk_send_restart(looper,
                           sdk_wallet_trustee,
                           sdk_pool_handle,
                           action=START,
                           datetime=str(datetime.isoformat(start_at)))

    _stopServer(server)
    for node in txnPoolNodeSet:
        assert node.restarter.lastActionEventInfo.ev_type == RestartLog.Events.scheduled
    _comparison_reply(responses, req_obj)


def test_restarter_can_initialize_after_pool_restart(txnPoolNodeSet):
    '''
    1. Add restart schedule message to ActionLog
    2. Add start restart message to ActionLog
    3. Check that Restarter can be create (emulate case after node restart).
    '''
    unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    restarted_node = txnPoolNodeSet[-1]
    ev_data = RestartLogData(unow)
    restarted_node.restarter._actionLog.append_scheduled(ev_data)
    restarted_node.restarter._actionLog.append_started(ev_data)
    Restarter(restarted_node.id,
              restarted_node.name,
              restarted_node.dataLocation,
              restarted_node.config,
              actionLog=restarted_node.restarter._actionLog)


def test_pool_restart_cancel(
        sdk_pool_handle, sdk_wallet_trustee, looper, tconf, txnPoolNodeSet):
    loop = asyncio.get_event_loop()
    server, indicator = loop.run_until_complete(
        _createServer(
            host=tconf.controlServiceHost,
            port=tconf.controlServicePort
        )
    )

    unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    start_at = unow + timedelta(seconds=1000)
    req_obj, responses = sdk_send_restart(looper,
                                          sdk_wallet_trustee,
                                          sdk_pool_handle,
                                          action=START,
                                          datetime=str(datetime.isoformat(start_at)))
    for node in txnPoolNodeSet:
        assert node.restarter.lastActionEventInfo.ev_type == RestartLog.Events.scheduled
    _comparison_reply(responses, req_obj)

    req_obj, responses = sdk_send_restart(looper,
                                          sdk_wallet_trustee,
                                          sdk_pool_handle,
                                          action=CANCEL,
                                          datetime="")
    _stopServer(server)
    for node in txnPoolNodeSet:
        assert node.restarter.lastActionEventInfo.ev_type == RestartLog.Events.cancelled
    _comparison_reply(responses, req_obj)


def test_pool_restart_now_without_datetime(
        sdk_pool_handle, sdk_wallet_trustee, looper, tdir, tconf, txnPoolNodeSet):
    pool_restart_now(sdk_pool_handle, sdk_wallet_trustee, looper,
                     tdir, tconf, START, txnPoolNodeSet)


def test_pool_restart_in_view_change(sdk_pool_handle, sdk_wallet_trustee, looper,
                                     tdir, tconf, txnPoolNodeSet):

    for node in txnPoolNodeSet:
        node.master_replica._consensus_data.waiting_for_new_view = True

    pool_restart_now(sdk_pool_handle, sdk_wallet_trustee, looper,
                     tdir, tconf, START, txnPoolNodeSet)


def pool_restart_now(sdk_pool_handle, sdk_wallet_trustee, looper, tdir, tconf,
                     action, txnPoolNodeSet, use_time=True):
    server, indicator = looper.loop.run_until_complete(
        _createServer(
            host=tconf.controlServiceHost,
            port=tconf.controlServicePort
        )
    )

    time = None
    if use_time:
        unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
        time = str(datetime.isoformat(unow + timedelta(seconds=1000)))
    req_obj, resp = sdk_send_restart(looper,
                                     sdk_wallet_trustee,
                                     sdk_pool_handle,
                                     action=action,
                                     datetime=time)
    _stopServer(server)
    _comparison_reply(resp, req_obj)


def _check_restart_log(item, action, when=None):
    assert item.ev_type == action and (
            when is None or str(datetime.isoformat(item.data.when)) == when)


def _comparison_reply(responses, req_obj):
    for json_resp in responses.values():
        resp = json.loads(json_resp)
        assert resp["op"] == REPLY
        assert resp[f.RESULT.nm][f.IDENTIFIER.nm] == req_obj[f.IDENTIFIER.nm]
        assert resp[f.RESULT.nm][f.REQ_ID.nm] == req_obj[f.REQ_ID.nm]
        assert resp[f.RESULT.nm][ACTION]
        assert resp[f.RESULT.nm][TXN_TYPE] == POOL_RESTART



from abc import ABCMeta
from logging import getLogger

import os

from indy_common.authorize.auth_actions import AbstractAuthAction
from indy_common.authorize.auth_constraints import AbstractAuthConstraint, AuthConstraint, \
    AuthConstraintAnd, ConstraintsEnum, AuthConstraintOr, AuthConstraintForbidden
from indy_common.authorize.helper import get_named_role
from indy_common.constants import ENDORSER, ENDORSER_STRING
from indy_common.roles import Roles
from indy_common.transactions import IndyTransactions
from indy_common.types import Request
from indy_node.persistence.idr_cache import IdrCache
from plenum.common.constants import STEWARD, TRUSTEE, TRUSTEE_STRING, STEWARD_STRING

logger = getLogger()


class AuthValidationError(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return "Validation error with reason: {}".format(self.reason)


class AbstractAuthorizer(metaclass=ABCMeta):

    def __init__(self):
        self.parent = None

    def authorize(self,
                  request: Request,
                  auth_constraint: AbstractAuthConstraint,
                  auth_action: AbstractAuthAction) -> (bool, str):
        raise NotImplementedError()


class RolesAuthorizer(AbstractAuthorizer):

    def __init__(self, cache: IdrCache):
        super().__init__()
        self.cache = cache

    def get_role(self, request: Request):
        """
        None roles are stored as empty strings, so the role returned as None
        by this function means that corresponding DID is not stored in a ledger.
        """
        idr = request.identifier
        if idr is None:
            return None
        return self._get_role(idr)

    def get_sig_count(self, request: Request, role: str = "*", off_ledger_signature=False):
        if request.signature:
            return 1 \
                if off_ledger_signature or self.is_role_accepted(self._get_role(request.identifier), role) \
                else 0
        elif not request.signatures:
            return 0
        if role == "*":
            return len(request.signatures)

        sig_count = 0
        for identifier, _ in request.signatures.items():
            signer_role = self._get_role(identifier)
            if off_ledger_signature or self.is_role_accepted(signer_role, role):
                sig_count += 1
        return sig_count

    def is_owner_accepted(self, constraint: AuthConstraint, action: AbstractAuthAction):
        if constraint.need_to_be_owner and not action.is_owner:
            return False
        return True

    def is_role_accepted(self, role, auth_constraint_role):
        # The field 'role' contains a value from the cache.
        # None - the identifier doesn't found in the cache.
        # "" - IDENTITY_OWNER
        # The field 'auth_constraint_role' contains a role from the auth_map
        # None - IDENTITY_OWNER
        if role is None:
            return False
        if role == "":
            role = None
        return role == auth_constraint_role or auth_constraint_role == '*'

    def is_sig_count_accepted(self, request: Request, auth_constraint: AuthConstraint):
        role = auth_constraint.role
        sig_count = self.get_sig_count(request, role=role, off_ledger_signature=auth_constraint.off_ledger_signature)
        return sig_count >= auth_constraint.sig_count

    def get_named_role_from_req(self, request: Request):
        return get_named_role(self.get_role(request))

    def authorize(self,
                  request: Request,
                  auth_constraint: AuthConstraint,
                  auth_action: AbstractAuthAction = None):
        # 1. Check that the Author is the owner
        # do first since it doesn't require going to state
        if not self.is_owner_accepted(auth_constraint, auth_action):
            if auth_action.field != '*':
                return False, "{} can not touch {} field since only the owner can modify it". \
                    format(self.get_named_role_from_req(request),
                           auth_action.field)
            else:
                return False, "{} can not edit {} txn since only owner can modify it". \
                    format(self.get_named_role_from_req(request),
                           IndyTransactions.get_name_from_code(auth_action.txn_type))

        author_role = self.get_role(request)

        # 2. Check that the Author is present on the ledger
        if auth_constraint.sig_count > 0 and not auth_constraint.off_ledger_signature and author_role is None:
            return False, "sender's DID {} is not found in the Ledger".format(request.identifier)

        # 3. Check that the Author signed the transaction in case of multi-sig
        if auth_constraint.sig_count > 0 and request.signatures and request.identifier not in request.signatures:
            return False, "Author must sign the transaction"

        # 4. Check that there are enough signatures of the needed role
        if not self.is_sig_count_accepted(request, auth_constraint):
            role = Roles(auth_constraint.role).name if auth_constraint.role != '*' else '*'
            return False, "Not enough {} signatures".format(role)

        return True, ""

    def _get_role(self, idr):
        try:
            return self.cache.getRole(idr, isCommitted=False)
        except KeyError:
            return None


class CompositeAuthorizer(AbstractAuthorizer):
    def __init__(self):
        super().__init__()
        self.authorizers = {}

    def register_authorizer(self, authorizer, auth_constraint_id=ConstraintsEnum.ROLE_CONSTRAINT_ID):
        authorizer.parent = self
        self.authorizers.setdefault(auth_constraint_id, []).append(authorizer)

    def authorize(self,
                  request: Request,
                  auth_constraint: AuthConstraint,
                  auth_action: AbstractAuthAction):
        for authorizer in self.authorizers.get(auth_constraint.constraint_id):
            authorized, reason = authorizer.authorize(request=request,
                                                      auth_constraint=auth_constraint,
                                                      auth_action=auth_action)
            if not authorized:
                raise AuthValidationError(reason)
        return True, ""


class AndAuthorizer(AbstractAuthorizer):

    def authorize(self,
                  request,
                  auth_constraint: AuthConstraintAnd,
                  auth_action: AbstractAuthAction):
        for constraint in auth_constraint.auth_constraints:
            authorized, reason = self.parent.authorize(request=request,
                                                       auth_constraint=constraint,
                                                       auth_action=auth_action)
            if not authorized:
                raise AuthValidationError(reason)
        return True, ""


class OrAuthorizer(AbstractAuthorizer):

    def authorize(self,
                  request: Request,
                  auth_constraint: AuthConstraintOr,
                  auth_action: AbstractAuthAction):
        successes = []
        fails = []
        for constraint in auth_constraint.auth_constraints:
            try:
                self.parent.authorize(request=request,
                                      auth_constraint=constraint,
                                      auth_action=auth_action)
            except AuthValidationError as e:
                logger.trace(e)
                fails.append("Constraint: {}, Error: {}".format(constraint, e.reason))
            else:
                successes.append(True)
        if len(successes) == 0:
            raise AuthValidationError(
                os.linesep.join(["Rule for this action is: {}".format(auth_constraint),
                                 "Failed checks:",
                                 os.linesep.join(fails)])
            )
        return True, ""


class ForbiddenAuthorizer(AbstractAuthorizer):

    def authorize(self,
                  request: Request,
                  auth_constraint: AuthConstraintForbidden,
                  auth_action: AbstractAuthAction):
        return False, str(auth_constraint)


class EndorserAuthorizer(AbstractAuthorizer):
    NO_NEED_FOR_ENDORSER_ROLES = {ENDORSER, TRUSTEE, STEWARD}
    NO_NEED_FOR_ENDORSER_ROLES_STR = [ENDORSER_STRING, TRUSTEE_STRING, STEWARD_STRING]

    ENDORSER_ROLES = {ENDORSER}
    ENDORSER_ROLES_STR = {ENDORSER_STRING}

    def __init__(self, cache: IdrCache):
        super().__init__()
        self.cache = cache

    def authorize(self,
                  request: Request,
                  auth_constraint: AuthConstraint,
                  auth_action: AbstractAuthAction = None):
        # check if endorser role is valid first
        res, reason = self._check_endorser_role(request)
        if not res:
            return res, reason

        # check if Endorser field must be explicitly present
        res, reason = self._check_endorser_field_presence(request)
        if not res:
            return res, reason

        return True, ""

    def _check_endorser_role(self, request):
        if request.endorser is None:
            return True, ""

        endorser_role = self._get_role(request.endorser)
        if endorser_role not in self.ENDORSER_ROLES:
            return False, "Endorser must have one of the following roles: {}".format(str(self.ENDORSER_ROLES_STR))

        return True, ""

    def _check_endorser_field_presence(self, request):
        # 1. Check that if Endorser is present, Endorser must sign the transaction
        # if author=endorser, then no multi-sig is required, since he's already signed the txn
        if request.endorser != request.identifier:
            if request.endorser and not request.signatures:
                return False, "Endorser must sign the transaction"
            if request.endorser and request.endorser not in request.signatures:
                return False, "Endorser must sign the transaction"

        # 2. Endorser is required only when the transaction is endorsed, that is signed by someone else besides the author.
        # If the auth rule requires an endorser to submit the transaction, then it will be caught by the Roles Authorizer,
        # so no need to check anything here if we don't have any extra signatures.
        if not request.signatures:
            return True, ""
        has_extra_sigs = len(request.signatures) > 1 or (request.identifier not in request.signatures)
        if not has_extra_sigs:
            return True, ""

        # 3. Endorser is required for unprivileged roles only
        author_role = self._get_role(request.identifier)
        if author_role == "":  # "" - IDENTITY_OWNER
            author_role = None
        if author_role in self.NO_NEED_FOR_ENDORSER_ROLES:
            return True, ""

        # 4. Endorser field is not required when multi-signed by non-privileged roles only
        has_privileged_sig = False
        for idr, _ in request.signatures.items():
            role = self._get_role(idr)
            if role in self.NO_NEED_FOR_ENDORSER_ROLES:
                has_privileged_sig = True
                break
        if not has_privileged_sig:
            return True, ""

        # 5. Check that Endorser field is present
        if request.endorser is None:
            return False, "'Endorser' field must be explicitly set for the endorsed transaction"

        return True, ""

    def _get_role(self, idr):
        if idr is None:
            return None
        try:
            return self.cache.getRole(idr, isCommitted=False)
        except KeyError:
            return None



from indy_common.authorize.auth_actions import compile_action_id, ADD_PREFIX, EDIT_PREFIX, split_action_id, ActionDef


def test_get_action_id_action_add(action_add):
    assert action_add.get_action_id() == compile_action_id(txn_type='SomeType',
                                                           field='some_field',
                                                           old_value='*',
                                                           new_value='new_value',
                                                           prefix=ADD_PREFIX)


def test_get_action_id_action_edit(action_edit):
    assert action_edit.get_action_id() == compile_action_id(txn_type='SomeType',
                                                            field='some_field',
                                                            old_value='old_value',
                                                            new_value='new_value',
                                                            prefix=EDIT_PREFIX)


def test_split_action_id():
    origin = ActionDef('SomeType', 'PREFIX', 'some_field', 'old_value', 'new_value')
    splitted_action = split_action_id(compile_action_id(txn_type='SomeType',
                                                        field='some_field',
                                                        old_value='old_value',
                                                        new_value='new_value',
                                                        prefix='PREFIX'))
    assert origin.prefix == splitted_action.prefix
    assert origin.txn_type == splitted_action.txn_type
    assert origin.field == splitted_action.field
    assert origin.old_value == splitted_action.old_value
    assert origin.new_value == splitted_action.new_value



from common.serializers.serialization import domain_state_serializer
from indy_common.authorize.auth_constraints import ConstraintsSerializer
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.state import config
from indy_common.types import ClientGetAuthRuleOperation
from indy_node.server.request_handlers.config_req_handlers.auth_rule.abstract_auth_rule_handler import \
    AbstractAuthRuleHandler
from indy_node.server.request_handlers.config_req_handlers.auth_rule.static_auth_rule_helper import StaticAuthRuleHelper
from plenum.common.txn_util import get_request_data
from plenum.server.request_handlers.handler_interfaces.read_request_handler import ReadRequestHandler

from indy_common.authorize.auth_actions import EDIT_PREFIX, split_action_id
from indy_common.constants import CONFIG_LEDGER_ID, AUTH_ACTION, OLD_VALUE, \
    NEW_VALUE, AUTH_TYPE, FIELD, CONSTRAINT, GET_AUTH_RULE
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager


class GetAuthRuleHandler(ReadRequestHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        self.write_req_validator = write_req_validator
        self.constraint_serializer = ConstraintsSerializer(domain_state_serializer)
        super().__init__(database_manager, GET_AUTH_RULE, CONFIG_LEDGER_ID)

    def static_validation(self, request: Request):
        self._validate_request_type(request)
        identifier, req_id, operation = get_request_data(request)

        required_fields = list(dict(ClientGetAuthRuleOperation.schema).keys())
        required_fields.remove(OLD_VALUE)
        if len(operation) > 1:
            if not set(required_fields).issubset(set(operation.keys())):
                raise InvalidClientRequest(identifier, req_id,
                                           "Not enough fields to build an auth key.")
            StaticAuthRuleHelper.check_auth_key(operation, identifier, req_id, self.write_req_validator.auth_map)

    def get_result(self, request: Request):
        self._validate_request_type(request)
        operation = request.operation
        proof = None
        if len(operation) >= len(ClientGetAuthRuleOperation.schema) - 1:
            path = StaticAuthRuleHelper.get_auth_key(operation)
            data, proof = self._get_auth_rule(path)
        else:
            data = self._get_all_auth_rules()
        result = self.make_result(request=request,
                                  data=data,
                                  proof=proof)
        result.update(operation)
        return result

    def _get_auth_rule(self, key):
        path = config.make_state_path_for_auth_rule(key)
        map_data, proof = self._get_value_from_state(path, with_proof=True)

        if map_data:
            data = self.constraint_serializer.deserialize(map_data)
        else:
            data = self.write_req_validator.auth_map[key]
        action_obj = split_action_id(key)
        return [self.make_auth_rule_data(data, action_obj)], proof

    def _get_all_auth_rules(self):
        data = self.write_req_validator.auth_map.copy()
        result = []
        for key in self.write_req_validator.auth_map:
            path = AbstractAuthRuleHandler.make_state_path_for_auth_rule(key)
            state_constraint, _ = self._get_value_from_state(path)
            if state_constraint:
                value = self.constraint_serializer.deserialize(state_constraint)
            else:
                value = data[key]
            action_obj = split_action_id(key)
            result.append(self.make_auth_rule_data(value, action_obj))
        return result

    @staticmethod
    def make_auth_rule_data(constraint, action_obj):
        result = {CONSTRAINT: constraint.as_dict if constraint is not None else {},
                  AUTH_TYPE: action_obj.txn_type,
                  AUTH_ACTION: action_obj.prefix,
                  FIELD: action_obj.field,
                  NEW_VALUE: action_obj.new_value,
                  }
        if action_obj.prefix == EDIT_PREFIX:
            result[OLD_VALUE] = action_obj.old_value
        return result



import json

from indy.did import create_and_store_my_did
from indy.ledger import build_get_validator_info_request

from indy_node.test.pool_restart.helper import sdk_get_and_check_multiply_replies
from plenum.common.util import randomString
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request


def sdk_get_validator_info(looper, steward_wallet, sdk_pool_handle, seed=None):
    # filling validator_info request and getting steward did
    seed = seed or randomString(32)
    validator_info_request, did = looper.loop.run_until_complete(
        prepare_validator_info_request(steward_wallet, seed))

    # sending request using 'sdk_' functions
    request_couple = sdk_sign_and_send_prepared_request(looper, steward_wallet,
                                                        sdk_pool_handle,
                                                        validator_info_request)
    # waiting for replies
    return sdk_get_and_check_multiply_replies(looper, request_couple)


async def prepare_validator_info_request(wallet, named_seed):
    wh, submitter_did = wallet
    (named_did, named_verkey) = \
        await create_and_store_my_did(wh, json.dumps({'seed': named_seed}))
    restart_request = await build_get_validator_info_request(submitter_did)
    return restart_request, named_did


from typing import Optional

from common.serializers.serialization import config_state_serializer
from indy_node.server.request_handlers.config_req_handlers.txn_author_agreement_handler import TxnAuthorAgreementHandler
from plenum.common.constants import TXN_AUTHOR_AGREEMENT, CONFIG_LEDGER_ID, TXN_AUTHOR_AGREEMENT_VERSION, \
    TXN_AUTHOR_AGREEMENT_TEXT
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.request import Request
from plenum.common.txn_util import get_payload_data, get_seq_no, get_txn_time
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler
from plenum.server.request_handlers.static_taa_helper import StaticTAAHelper
from plenum.server.request_handlers.utils import encode_state_value, decode_state_value


class TxnAuthorAgreementHandlerV1(TxnAuthorAgreementHandler):
    def dynamic_validation(self, request: Request, req_pp_time: Optional[int]):
        # TODO: Have real implementation here?
        pass

    def update_state(self, txn, prev_result, request, is_committed=False):
        self._validate_txn_type(txn)
        payload = get_payload_data(txn)
        text = payload.get(TXN_AUTHOR_AGREEMENT_TEXT)
        version = payload[TXN_AUTHOR_AGREEMENT_VERSION]
        seq_no = get_seq_no(txn)
        txn_time = get_txn_time(txn)
        self._update_txn_author_agreement(seq_no, txn_time, text, version)

    def _update_txn_author_agreement(self, seq_no, txn_time, text, version):
        digest = StaticTAAHelper.taa_digest(text, version)
        data = encode_state_value({
            TXN_AUTHOR_AGREEMENT_TEXT: text,
            TXN_AUTHOR_AGREEMENT_VERSION: version
        }, seq_no, txn_time, serializer=config_state_serializer)

        self.state.set(StaticTAAHelper.state_path_taa_digest(digest), data)
        self.state.set(StaticTAAHelper.state_path_taa_latest(), digest)
        self.state.set(StaticTAAHelper.state_path_taa_version(version), digest)



from indy_common.types import SafeRequest
from plenum.common.constants import CURRENT_PROTOCOL_VERSION, TXN_TYPE, TARGET_NYM, VERKEY, NYM
from plenum.common.types import f, OPERATION
from plenum.test.input_validation.constants import TEST_TARGET_NYM, TEST_VERKEY_ABBREVIATED


def test_client_safe_req_not_strict_by_default():
    operation = {
        TXN_TYPE: NYM,
        TARGET_NYM: TEST_TARGET_NYM,
        VERKEY: TEST_VERKEY_ABBREVIATED,
        "some_new_field_op1": "some_new_value_op1",
        "some_new_field_op2": "some_new_value_op2"
    }
    kwargs = {f.IDENTIFIER.nm: "1" * 16,
              f.REQ_ID.nm: 1,
              OPERATION: operation,
              f.SIG.nm: "signature",
              f.PROTOCOL_VERSION.nm: CURRENT_PROTOCOL_VERSION,
              "some_new_field1": "some_new_value1",
              "some_new_field2": "some_new_value2"}
    assert SafeRequest(**kwargs)



import pytest

from indy_common.constants import CRED_DEF_ID
from indy_common.types import Request
from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.util import randomString


def test_validation_cred_def_not_present(build_revoc_def_by_default,
                                         create_node_and_not_start):
    node = create_node_and_not_start
    req = build_revoc_def_by_default
    with pytest.raises(InvalidClientRequest, match="There is no any CRED_DEF"):
        node.write_manager.dynamic_validation(Request(**req), 0)


def test_invalid_cred_def_id_format(build_revoc_def_by_default,
                                    create_node_and_not_start):
    node = create_node_and_not_start
    req = build_revoc_def_by_default

    req['operation'][CRED_DEF_ID] = ":".join(3 * [randomString(10)])
    with pytest.raises(InvalidClientRequest, match="Format of {} field is not acceptable".format(CRED_DEF_ID)):
        node.write_manager.static_validation(Request(**req))

    req['operation'][CRED_DEF_ID] = ":".join(6 * [randomString(10)])
    with pytest.raises(InvalidClientRequest, match="Format of {} field is not acceptable".format(CRED_DEF_ID)):
        node.write_manager.static_validation(Request(**req))



import pytest
from plenum.common.constants import CURRENT_PROTOCOL_VERSION

from indy_common.req_utils import get_write_schema_name, get_write_schema_version, get_write_schema_attr_names, \
    get_read_schema_version, get_read_schema_name, get_read_schema_from, get_txn_schema_name, get_txn_schema_version, \
    get_txn_schema_attr_names, get_reply_schema_version, get_reply_schema_name, get_reply_schema_from, \
    get_reply_schema_attr_names
from indy_common.types import SafeRequest
from plenum.common.txn_util import reqToTxn


@pytest.fixture(scope="module")
def write_schema_request(request):
    req = {
        'operation': {
            'type': '101',
            'data': {
                'version': '1.0',
                'name': 'Degree',
                'attr_names': ['undergrad', 'last_name', 'first_name', 'birth_date', 'postgrad', 'expiry_date']
            },
        },

        'identifier': 'L5AD5g65TDQr1PPHHRoiGf',
        'reqId': 1514280215504647,
        'protocolVersion': CURRENT_PROTOCOL_VERSION,
        'signature': '5ZTp9g4SP6t73rH2s8zgmtqdXyTuSMWwkLvfV1FD6ddHCpwTY5SAsp8YmLWnTgDnPXfJue3vJBWjy89bSHvyMSdS'
    }
    return SafeRequest(**req)


@pytest.fixture(scope="module")
def read_schema_request():
    req = {
        'operation': {
            'type': '107',
            'dest': 'L5AD5g65TDQr1PPHHRoiGf',
            'data': {
                'version': '1.0',
                'name': 'Degree',
            }
        },

        'identifier': 'E5AD5g65TDQr1PPHHRoiGf',
        'reqId': 1514280215504647,
        'protocolVersion': CURRENT_PROTOCOL_VERSION
    }
    return SafeRequest(**req)


@pytest.fixture(scope="module")
def get_schema_reply():
    return {
        'type': '107',
        'dest': 'L5AD5g65TDQr1PPHHRoiGf',
        'data': {
            'version': '1.0',
            'name': 'Degree',
            'attr_names': ['undergrad', 'last_name', 'first_name', 'birth_date', 'postgrad', 'expiry_date']
        },

        'identifier': 'E5AD5g65TDQr1PPHHRoiGf',
        'reqId': 1514280215504647,
        'protocolVersion': CURRENT_PROTOCOL_VERSION,

        'seqNo': 10,
        'txnTime': 1514214795,
    }


def test_get_write_schema_name(write_schema_request):
    assert 'Degree' == get_write_schema_name(write_schema_request)


def test_get_write_schema_version(write_schema_request):
    assert '1.0' == get_write_schema_version(write_schema_request)


def test_get_write_schema_attr_names(write_schema_request):
    assert ['undergrad', 'last_name', 'first_name', 'birth_date', 'postgrad',
            'expiry_date'] == get_write_schema_attr_names(write_schema_request)


def test_get_txn_schema_name(write_schema_request):
    txn = reqToTxn(write_schema_request)
    assert 'Degree' == get_txn_schema_name(txn)


def test_get_txn_schema_version(write_schema_request):
    txn = reqToTxn(write_schema_request)
    assert '1.0' == get_txn_schema_version(txn)


def test_get_txn_schema_attr_names(write_schema_request):
    txn = reqToTxn(write_schema_request)
    assert ['undergrad', 'last_name', 'first_name', 'birth_date', 'postgrad',
            'expiry_date'] == get_txn_schema_attr_names(txn)


def test_get_read_schema_name(read_schema_request):
    assert 'Degree' == get_read_schema_name(read_schema_request)


def test_get_read_schema_version(read_schema_request):
    assert '1.0' == get_read_schema_version(read_schema_request)


def test_get_read_schema_from(read_schema_request):
    assert 'L5AD5g65TDQr1PPHHRoiGf' == get_read_schema_from(read_schema_request)


def test_get_reply_schema_name(get_schema_reply):
    assert 'Degree' == get_reply_schema_name(get_schema_reply)


def test_get_reply_schema_version(get_schema_reply):
    assert '1.0' == get_reply_schema_version(get_schema_reply)


def test_get_reply_schema_from(get_schema_reply):
    assert 'L5AD5g65TDQr1PPHHRoiGf' == get_reply_schema_from(get_schema_reply)


def test_get_reply_schema_attr_names(get_schema_reply):
    assert ['undergrad', 'last_name', 'first_name', 'birth_date', 'postgrad',
            'expiry_date'] == get_reply_schema_attr_names(get_schema_reply)



import json

import pytest
from indy.ledger import build_get_schema_request
from plenum.common.exceptions import RequestNackedException

from plenum.common.constants import DATA, NAME, VERSION, TXN_METADATA, TXN_METADATA_SEQ_NO

from plenum.common.types import OPERATION

from plenum.test.helper import sdk_sign_and_submit_req, sdk_get_and_check_replies

from indy_node.test.api.helper import sdk_write_schema
from indy_node.test.helper import createUuidIdentifier, modify_field


@pytest.fixture(scope="module")
def send_schema(looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee):
    schema_json, _ = sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_trustee)
    return json.loads(schema_json)['id']


@pytest.fixture(scope="module")
def send_schema_seq_no(looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee):
    _, reply = sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_trustee)
    return reply['result'][TXN_METADATA][TXN_METADATA_SEQ_NO]


@pytest.fixture(scope="module")
def send_schema_req(looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee):
    schema_json, reply = sdk_write_schema(
        looper, sdk_pool_handle,
        sdk_wallet_trustee,
        ["attrib1", "attrib2"],
        "some_name_ajf",
        "1.0")
    return schema_json, reply


def test_send_get_schema_succeeds(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, send_schema):
    _, did = sdk_wallet_trustee

    request = looper.loop.run_until_complete(build_get_schema_request(did, send_schema))
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    assert rep[0][1]['result']['seqNo']


def test_send_get_schema_as_client(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_client, send_schema):
    _, did = sdk_wallet_client

    request = looper.loop.run_until_complete(build_get_schema_request(did, send_schema))
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, request)])
    assert rep[0][1]['result']['seqNo']


def test_send_get_schema_fails_with_invalid_name(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, send_schema):
    _, did = sdk_wallet_trustee

    request = looper.loop.run_until_complete(build_get_schema_request(did, send_schema))
    request = modify_field(request, 'name111', OPERATION, DATA, NAME)
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    assert rep[0][1]['result']['seqNo'] is None


def test_send_get_schema_fails_with_invalid_dest(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, send_schema):
    uuid_identifier = createUuidIdentifier()
    _, did = sdk_wallet_trustee

    request = looper.loop.run_until_complete(build_get_schema_request(did, send_schema))
    request = modify_field(request, uuid_identifier, OPERATION, 'dest')
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    assert rep[0][1]['result']['seqNo'] is None


def test_send_get_schema_fails_with_invalid_version(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, send_schema):
    _, did = sdk_wallet_trustee

    request = looper.loop.run_until_complete(build_get_schema_request(did, send_schema))
    request = modify_field(request, '2.0', OPERATION, DATA, VERSION)
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    assert rep[0][1]['result']['seqNo'] is None


def test_send_get_schema_fails_with_invalid_version_syntax(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, send_schema):
    _, did = sdk_wallet_trustee

    request = looper.loop.run_until_complete(build_get_schema_request(did, send_schema))
    request = modify_field(request, 'asd', OPERATION, DATA, VERSION)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    e.match("Invalid version: 'asd'")


def test_send_get_schema_fails_without_version(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, send_schema):
    _, did = sdk_wallet_trustee

    request = looper.loop.run_until_complete(build_get_schema_request(did, send_schema))
    request = json.loads(request)
    del request[OPERATION][DATA][VERSION]
    request = json.dumps(request)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    e.match('missed fields - version')


def test_send_get_schema_fails_without_name(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, send_schema):
    _, did = sdk_wallet_trustee

    request = looper.loop.run_until_complete(build_get_schema_request(did, send_schema))
    request = json.loads(request)
    del request[OPERATION][DATA][NAME]
    request = json.dumps(request)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    e.match('missed fields - name')


def test_send_get_schema_fails_without_dest(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, send_schema):
    _, did = sdk_wallet_trustee

    request = looper.loop.run_until_complete(build_get_schema_request(did, send_schema))
    request = json.loads(request)
    del request[OPERATION]['dest']
    request = json.dumps(request)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    e.match('missed fields - dest')



from copy import deepcopy

from indy_node.test import waits
from stp_core.loop.eventually import eventually
from plenum.common.constants import VERSION
from indy_common.constants import REINSTALL

from indy_node.test.upgrade.helper import bumpedVersion, checkUpgradeScheduled, \
    check_no_loop, sdk_ensure_upgrade_sent, clear_aq_stash
from indy_node.server.upgrade_log import UpgradeLog

whitelist = ['Failed to upgrade node',
             'failed upgrade',
             'This problem may have external reasons, check syslog for more information']


def test_upgrade_does_not_get_into_loop_if_reinstall_and_failed(
        looper, tconf, nodeSet, validUpgrade, sdk_pool_handle, sdk_wallet_trustee, monkeypatch):
    new_version = bumpedVersion(validUpgrade['version'])
    upgr1 = deepcopy(validUpgrade)
    upgr1[VERSION] = new_version
    upgr1[REINSTALL] = True

    clear_aq_stash(nodeSet)

    # An upgrade scheduled, it should pass
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgr1)
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            upgr1[VERSION],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))

    # we have not patched indy_node version so nodes think the upgrade had
    # failed
    check_no_loop(nodeSet, UpgradeLog.Events.failed)



import json

import pytest

from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.authorize.auth_constraints import ROLE, SIG_COUNT
from indy_common.constants import AUTH_ACTION, OLD_VALUE, CONSTRAINT, NEW_VALUE, NYM, ENDORSER, GET_AUTH_RULE
from indy_node.test.auth_rule.helper import sdk_send_and_check_get_auth_rule_request, \
    sdk_send_and_check_auth_rules_request_invalid
from indy_node.test.helper import sdk_send_and_check_req_json, sdk_send_and_check_auth_rules_request, generate_auth_rule
from plenum.common.constants import TRUSTEE, STEWARD, DATA, TXN_TYPE
from plenum.common.exceptions import RequestRejectedException, \
    RequestNackedException
from plenum.test.helper import sdk_gen_request

RESULT = "result"


def _send_and_check_get_auth_rule(looper, sdk_pool_handle, sdk_wallet, key):
    op = {TXN_TYPE: GET_AUTH_RULE,
          **key}
    req_obj = sdk_gen_request(op, identifier=sdk_wallet[1])
    return sdk_send_and_check_req_json(
        looper, sdk_pool_handle, sdk_wallet, json.dumps(req_obj.as_dict)
    )


def test_auth_rules_transaction_without_changes(looper,
                                                sdk_wallet_trustee,
                                                sdk_pool_handle):
    before_resp = sdk_send_and_check_get_auth_rule_request(looper,
                                                           sdk_pool_handle,
                                                           sdk_wallet_trustee)
    sdk_send_and_check_auth_rules_request(looper,
                                          sdk_pool_handle,
                                          sdk_wallet_trustee,
                                          rules=before_resp[0][1][RESULT][DATA])
    after_resp = sdk_send_and_check_get_auth_rule_request(looper,
                                                          sdk_pool_handle,
                                                          sdk_wallet_trustee)
    assert before_resp[0][1]["result"][DATA] == after_resp[0][1]["result"][DATA]


def test_auth_rules_transaction(looper,
                                sdk_wallet_trustee,
                                sdk_pool_handle):
    rule = generate_auth_rule()
    key = dict(rule)
    key.pop(CONSTRAINT)
    sdk_send_and_check_auth_rules_request(looper,
                                          sdk_pool_handle,
                                          sdk_wallet_trustee,
                                          rules=[rule])
    after_resp = _send_and_check_get_auth_rule(looper,
                                               sdk_pool_handle,
                                               sdk_wallet_trustee,
                                               key)
    assert [rule] == after_resp[0][1]["result"][DATA]


def test_reject_all_rules_from_auth_rules_txn(looper,
                                              sdk_wallet_trustee,
                                              sdk_pool_handle):
    _, before_resp = sdk_send_and_check_get_auth_rule_request(looper,
                                                              sdk_pool_handle,
                                                              sdk_wallet_trustee)[0]
    rules = [generate_auth_rule(ADD_PREFIX, NYM,
                                ROLE, "wrong_new_value"),
             generate_auth_rule(EDIT_PREFIX, NYM,
                                ROLE, ENDORSER, TRUSTEE)]
    with pytest.raises(RequestNackedException):
        sdk_send_and_check_auth_rules_request(looper,
                                              sdk_pool_handle,
                                              sdk_wallet_trustee,
                                              rules=rules)
    _, after_resp = sdk_send_and_check_get_auth_rule_request(looper,
                                                             sdk_pool_handle,
                                                             sdk_wallet_trustee)[0]
    assert before_resp["result"][DATA] == after_resp["result"][DATA]


def test_reject_with_empty_rules_list(looper,
                                      sdk_wallet_trustee,
                                      sdk_pool_handle):
    with pytest.raises(RequestNackedException,
                       match="client request invalid.*InvalidClientRequest.*length should be at least 1"):
        sdk_send_and_check_auth_rules_request_invalid(looper,
                                                      sdk_pool_handle,
                                                      sdk_wallet_trustee,
                                                      rules=[])


def test_reject_with_unacceptable_role_in_constraint(looper,
                                                     sdk_wallet_trustee,
                                                     sdk_pool_handle):
    rule = generate_auth_rule()
    unacceptable_role = 'olololo'
    rule[CONSTRAINT][ROLE] = unacceptable_role
    with pytest.raises(RequestNackedException) as e:
        sdk_send_and_check_auth_rules_request(looper,
                                              sdk_pool_handle,
                                              sdk_wallet_trustee,
                                              rules=[rule])
    e.match('InvalidClientRequest')
    e.match('client request invalid')
    e.match('Role {} is not acceptable'.format(unacceptable_role))


def test_reject_auth_rules_transaction(looper,
                                       sdk_wallet_steward,
                                       sdk_pool_handle):
    with pytest.raises(RequestRejectedException) as e:
        sdk_send_and_check_auth_rules_request(looper,
                                              sdk_pool_handle,
                                              sdk_wallet_steward)
    e.match('Not enough TRUSTEE signatures')


def test_reqnack_auth_rules_transaction_with_wrong_key(looper,
                                                       sdk_wallet_trustee,
                                                       sdk_pool_handle):
    with pytest.raises(RequestNackedException) as e:
        sdk_send_and_check_auth_rules_request(looper,
                                              sdk_pool_handle,
                                              sdk_wallet_trustee,
                                              [generate_auth_rule(auth_type="*")])
    e.match('InvalidClientRequest')
    e.match("client request invalid")
    e.match("is not found in authorization map")


def test_reqnack_auth_rules_edit_transaction_with_wrong_format(looper,
                                                               sdk_wallet_trustee,
                                                               sdk_pool_handle):
    rule = generate_auth_rule(auth_action=EDIT_PREFIX)
    rule.pop(OLD_VALUE)
    with pytest.raises(RequestNackedException) as e:
        sdk_send_and_check_auth_rules_request_invalid(looper,
                                                      sdk_pool_handle,
                                                      sdk_wallet_trustee,
                                                      rules=[rule])
    e.match('InvalidClientRequest')
    e.match("client request invalid")
    e.match("Transaction for change authentication "
            "rule for {}={} must contain field {}".
            format(AUTH_ACTION, EDIT_PREFIX, OLD_VALUE))


def test_reqnack_auth_rules_add_transaction_with_wrong_format(looper,
                                                              sdk_wallet_trustee,
                                                              sdk_pool_handle):
    with pytest.raises(RequestNackedException) as e:
        sdk_send_and_check_auth_rules_request_invalid(looper,
                                                      sdk_pool_handle,
                                                      sdk_wallet_trustee,
                                                      [generate_auth_rule(old_value="*")])
    e.match('InvalidClientRequest')
    e.match("client request invalid")
    e.match("Transaction for change authentication "
            "rule for {}={} must not contain field {}".
            format(AUTH_ACTION, ADD_PREFIX, OLD_VALUE))



from indy_common.authorize.auth_actions import AuthActionEdit, AuthActionAdd
from plenum.common.constants import SERVICES, NODE, NODE_IP, NODE_PORT, CLIENT_PORT, \
    CLIENT_IP, BLS_KEY, ALIAS


def test_node_enable(write_request_validation, req, is_owner):
    authorized = (req.identifier == "steward_identifier" and is_owner)
    assert authorized == write_request_validation(req,
                                                  [AuthActionAdd(txn_type=NODE,
                                                                 field=SERVICES,
                                                                 value='[\'VALIDATOR\']',
                                                                 is_owner=is_owner)])


def test_node_enable_with_empty_services(write_request_validation, req, is_owner):
    authorized = (req.identifier == "steward_identifier" and is_owner)
    assert authorized == write_request_validation(req,
                                                  [AuthActionAdd(txn_type=NODE,
                                                                 field=SERVICES,
                                                                 value=[],
                                                                 is_owner=is_owner)])


def test_node_promote(write_request_validation, req, is_owner):
    authorized = (req.identifier == "steward_identifier" and is_owner) or (req.identifier == "trustee_identifier")
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=NODE,
                                                                  field=SERVICES,
                                                                  old_value=[],
                                                                  new_value=['VALIDATOR'],
                                                                  is_owner=is_owner)])


def test_node_demote(write_request_validation, req, is_owner):
    authorized = (req.identifier == "steward_identifier" and is_owner) or (req.identifier == "trustee_identifier")
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=NODE,
                                                                  field=SERVICES,
                                                                  old_value=['VALIDATOR'],
                                                                  new_value=[],
                                                                  is_owner=is_owner)])


def test_node_wrong_old_service_name(write_request_validation, req, is_owner):
    assert not write_request_validation(req,
                                        [AuthActionEdit(txn_type=NODE,
                                                        field=SERVICES,
                                                        old_value='aaa',
                                                        new_value=[],
                                                        is_owner=is_owner)])


def test_node_wrong_new_service_name(write_request_validation, req, is_owner):
    assert not write_request_validation(req,
                                        [AuthActionEdit(txn_type=NODE,
                                                        field=SERVICES,
                                                        old_value=[],
                                                        new_value='aaa',
                                                        is_owner=is_owner)])


def test_node_change_node_ip(write_request_validation, req, is_owner):
    authorized = (req.identifier == "steward_identifier" and is_owner)
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=NODE,
                                                                  field=NODE_IP,
                                                                  old_value='old_value',
                                                                  new_value='new_value',
                                                                  is_owner=is_owner)])


def test_node_change_node_port(write_request_validation, req, is_owner):
    authorized = (req.identifier == "steward_identifier" and is_owner)
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=NODE,
                                                                  field=NODE_PORT,
                                                                  old_value='old_value',
                                                                  new_value='new_value',
                                                                  is_owner=is_owner)])


def test_node_change_client_ip(write_request_validation, req, is_owner):
    authorized = (req.identifier == "steward_identifier" and is_owner)
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=NODE,
                                                                  field=CLIENT_IP,
                                                                  old_value='old_value',
                                                                  new_value='new_value',
                                                                  is_owner=is_owner)])


def test_node_change_client_port(write_request_validation, req, is_owner):
    authorized = (req.identifier == "steward_identifier" and is_owner)
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=NODE,
                                                                  field=CLIENT_PORT,
                                                                  old_value='old_value',
                                                                  new_value='new_value',
                                                                  is_owner=is_owner)])


def test_node_change_bls_keys(write_request_validation, req, is_owner):
    authorized = (req.identifier == "steward_identifier" and is_owner)
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=NODE,
                                                                  field=BLS_KEY,
                                                                  old_value='old_value',
                                                                  new_value='new_value',
                                                                  is_owner=is_owner)])


def test_node_change_alias(write_request_validation, req, is_owner):
    authorized = False  # alias can not be changed
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=NODE,
                                                                  field=ALIAS,
                                                                  old_value='old_value',
                                                                  new_value='new_value',
                                                                  is_owner=is_owner)])



from plenum.common.constants import STATE_PROOF, ROOT_HASH, MULTI_SIGNATURE, PROOF_NODES, MULTI_SIGNATURE_SIGNATURE, \
    MULTI_SIGNATURE_PARTICIPANTS, MULTI_SIGNATURE_VALUE, MULTI_SIGNATURE_VALUE_LEDGER_ID, \
    MULTI_SIGNATURE_VALUE_STATE_ROOT, MULTI_SIGNATURE_VALUE_TXN_ROOT, MULTI_SIGNATURE_VALUE_POOL_STATE_ROOT, \
    MULTI_SIGNATURE_VALUE_TIMESTAMP
from plenum.test.helper import sdk_sign_and_submit_op, sdk_get_and_check_replies


def sdk_submit_operation_and_get_result(looper, sdk_pool_handle, sdk_wallet_sender, operation):
    req = sdk_sign_and_submit_op(looper, sdk_pool_handle, sdk_wallet_sender, operation)
    replies = sdk_get_and_check_replies(looper, [req])
    assert len(replies) == 1
    return replies[0][1]['result']


def check_valid_proof(result):
    assert STATE_PROOF in result

    state_proof = result[STATE_PROOF]
    assert ROOT_HASH in state_proof
    assert state_proof[ROOT_HASH]
    assert PROOF_NODES in state_proof
    assert state_proof[PROOF_NODES]
    assert MULTI_SIGNATURE in state_proof

    multi_sig = state_proof[MULTI_SIGNATURE]
    assert multi_sig
    assert multi_sig[MULTI_SIGNATURE_PARTICIPANTS]
    assert multi_sig[MULTI_SIGNATURE_SIGNATURE]
    assert MULTI_SIGNATURE_VALUE in multi_sig

    multi_sig_value = multi_sig[MULTI_SIGNATURE_VALUE]
    assert MULTI_SIGNATURE_VALUE_LEDGER_ID in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_LEDGER_ID]
    assert MULTI_SIGNATURE_VALUE_STATE_ROOT in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_STATE_ROOT]
    assert MULTI_SIGNATURE_VALUE_TXN_ROOT in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_TXN_ROOT]
    assert MULTI_SIGNATURE_VALUE_POOL_STATE_ROOT in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_POOL_STATE_ROOT]
    assert MULTI_SIGNATURE_VALUE_TIMESTAMP in multi_sig_value
    assert multi_sig_value[MULTI_SIGNATURE_VALUE_TIMESTAMP]



from contextlib import contextmanager
import json
from unittest import mock

from ledger.util import F
from plenum.common.constants import IDENTIFIER, STEWARD, TARGET_NYM, TXN_TYPE, VERKEY
from plenum.common.exceptions import InvalidClientRequest, UnauthorizedClientRequest
from plenum.common.request import Request
from plenum.common.txn_util import append_txn_metadata, reqToTxn
from plenum.common.util import randomString
from plenum.server.request_handlers.utils import nym_to_state_key
import pytest

from indy_common.auth import Authoriser
from indy_common.constants import (
    DIDDOC_CONTENT,
    NYM,
    NYM_VERSION,
    NYM_VERSION_CONVENTION,
    NYM_VERSION_SELF_CERT,
    ROLE,
)
from indy_node.server.request_handlers.domain_req_handlers.nym_handler import NymHandler
from indy_node.test.request_handlers.helper import add_to_idr, get_exception


@pytest.fixture(scope="module")
def nym_handler(db_manager, tconf, write_auth_req_validator):
    return NymHandler(tconf, db_manager, write_auth_req_validator)


@pytest.fixture(scope="module")
def creator(db_manager):
    identifier = randomString()
    idr = db_manager.idr_cache
    add_to_idr(idr, identifier, None)
    return identifier


@pytest.fixture
def doc():
    yield {
        "@context": [
            "https://www.w3.org/ns/did/v1",
            "https://identity.foundation/didcomm-messaging/service-endpoint/v1",
        ],
        "serviceEndpoint": [
            {
                "id": "did:indy:sovrin:123456#didcomm",
                "type": "didcomm-messaging",
                "serviceEndpoint": "https://example.com",
                "recipientKeys": ["#verkey"],
                "routingKeys": [],
            }
        ],
    }


@pytest.fixture
def nym_request_factory(creator, doc):
    def _factory(diddoc_content: dict = None, version: int = None):
        if diddoc_content is None:
            diddoc_content = doc
        return Request(
            identifier=creator,
            reqId=5,
            operation={
                TXN_TYPE: NYM,
                TARGET_NYM: "X3XUxYQM2cfkSMzfMNma73",
                ROLE: None,
                DIDDOC_CONTENT: json.dumps(diddoc_content),
                VERKEY: "HNjfjoeZ7WAHYDSzWcvzyvUABepctabD7QSxopM48fYx",
                **({NYM_VERSION: version} if version is not None else {})
            },
        )

    yield _factory


@pytest.fixture(scope="function")
def nym_request(nym_request_factory):
    yield nym_request_factory()


def test_nym_static_validation_passes(nym_request, nym_handler: NymHandler):
    nym_handler.static_validation(nym_request)


def test_nym_static_validation_failed_without_dest(
    nym_request, nym_handler: NymHandler
):
    del nym_request.operation["dest"]
    with pytest.raises(InvalidClientRequest):
        nym_handler.static_validation(nym_request)


def test_nym_static_validation_failed_with_none_dest(
    nym_request, nym_handler: NymHandler
):
    nym_request.operation["dest"] = None
    with pytest.raises(InvalidClientRequest):
        nym_handler.static_validation(nym_request)


def test_nym_static_validation_failed_with_empty_dest(
    nym_request, nym_handler: NymHandler
):
    nym_request.operation["dest"] = ""
    with pytest.raises(InvalidClientRequest):
        nym_handler.static_validation(nym_request)


def test_nym_static_validation_failed_with_spaced_dest(
    nym_request, nym_handler: NymHandler
):
    nym_request.operation["dest"] = " " * 5
    with pytest.raises(InvalidClientRequest):
        nym_handler.static_validation(nym_request)


def test_nym_static_validation_authorized(nym_request, nym_handler: NymHandler):
    for role in Authoriser.ValidRoles:
        nym_request.operation["role"] = role
        nym_handler.static_validation(nym_request)


def test_nym_static_validation_not_authorized_random(
    nym_request, nym_handler: NymHandler
):
    nym_request.operation["role"] = randomString()
    with pytest.raises(InvalidClientRequest):
        nym_handler.static_validation(nym_request)


def test_nym_static_validation_fails_diddoc_content_with_id(
    nym_request_factory, doc, nym_handler: NymHandler
):
    doc["id"] = randomString()
    nym_request = nym_request_factory(doc)
    with pytest.raises(InvalidClientRequest):
        nym_handler.static_validation(nym_request)


def test_nym_static_validation_diddoc_content_without_context(
    nym_request_factory, doc, nym_handler: NymHandler
):
    del doc["@context"]
    nym_request = nym_request_factory(doc)
    nym_handler.static_validation(nym_request)


def test_nym_static_validation_diddoc_content_fails_with_same_id(
    nym_request_factory, doc, nym_handler: NymHandler
):
    doc["authentication"] = [{
        "id": "did:indy:sovrin:123456#verkey"
    }]
    nym_request = nym_request_factory(doc)
    with pytest.raises(InvalidClientRequest):
        nym_handler.static_validation(nym_request)


def test_nym_dynamic_validation_for_new_nym(
    nym_request, nym_handler: NymHandler, creator
):
    nym_handler.write_req_validator.validate = get_exception(False)
    add_to_idr(nym_handler.database_manager.idr_cache, creator, STEWARD)
    nym_handler.dynamic_validation(nym_request, 0)

    nym_handler.write_req_validator.validate = get_exception(True)
    with pytest.raises(UnauthorizedClientRequest):
        nym_handler.dynamic_validation(nym_request, 0)


def test_nym_dynamic_validation_for_existing_nym(
    nym_request: Request, nym_handler: NymHandler, creator
):
    add_to_idr(
        nym_handler.database_manager.idr_cache, nym_request.operation["dest"], None
    )
    nym_handler.write_req_validator.validate = get_exception(False)
    add_to_idr(nym_handler.database_manager.idr_cache, creator, STEWARD)
    nym_handler.dynamic_validation(nym_request, 0)

    nym_handler.write_req_validator.validate = get_exception(True)
    with pytest.raises(UnauthorizedClientRequest):
        nym_handler.dynamic_validation(nym_request, 0)


def test_nym_dynamic_validation_for_existing_nym_fails_with_no_changes(
    nym_handler: NymHandler, creator
):
    nym_request = Request(
        identifier=creator, reqId=5, operation={"type": NYM, "dest": randomString()}
    )
    add_to_idr(
        nym_handler.database_manager.idr_cache, nym_request.operation["dest"], None
    )
    add_to_idr(nym_handler.database_manager.idr_cache, creator, STEWARD)

    nym_handler.write_req_validator.validate = get_exception(True)
    with pytest.raises(InvalidClientRequest):
        nym_handler.dynamic_validation(nym_request, 0)


def test_update_state(nym_request: Request, nym_handler: NymHandler):
    seq_no = 1
    txn_time = 1560241033
    nym = nym_request.operation.get(TARGET_NYM)
    txn = reqToTxn(nym_request)
    append_txn_metadata(txn, seq_no, txn_time)
    path = nym_to_state_key(nym)

    result = nym_handler.update_state(txn, None, nym_request)
    state_value = nym_handler.get_from_state(path)
    assert state_value == result
    assert state_value[IDENTIFIER] == nym_request.identifier
    assert state_value[F.seqNo.name] == seq_no
    assert state_value[ROLE] == nym_request.operation.get(ROLE)
    assert state_value[VERKEY] == nym_request.operation.get(VERKEY)
    assert state_value[DIDDOC_CONTENT] == nym_request.operation.get(DIDDOC_CONTENT)


def test_fail_on_version_update(nym_request: Request, nym_handler: NymHandler, doc, nym_request_factory):
    nym_request = nym_request_factory(doc, NYM_VERSION_SELF_CERT)
    txn = reqToTxn(nym_request)
    seq_no = 1
    txn_time = 1560241033
    append_txn_metadata(txn, seq_no, txn_time)
    nym_data = nym_handler.update_state(txn, None, None, False)
    with pytest.raises(InvalidClientRequest) as e:
        nym_handler._validate_existing_nym(nym_request, nym_request.operation, nym_data)
    e.match("Cannot set version on existing nym")


@pytest.mark.parametrize("version", [-1, 3, 100])
def test_fail_on_bad_version(version: int, nym_request: Request, nym_handler: NymHandler, doc, nym_request_factory):
    nym_request = nym_request_factory(doc, version)
    txn = reqToTxn(nym_request)
    seq_no = 1
    txn_time = 1560241033
    append_txn_metadata(txn, seq_no, txn_time)
    nym_data = nym_handler.update_state(txn, None, None, False)
    with pytest.raises(InvalidClientRequest) as e:
        nym_handler.static_validation(nym_request)
    e.match("Version must be one of")


def test_nym_validation_legacy_convention_x(nym_request: Request, nym_handler: NymHandler, doc, nym_request_factory):
    nym_request = nym_request_factory(doc, NYM_VERSION_CONVENTION)
    nym_request.operation[TARGET_NYM] = "NOTDERIVEDFROMVERKEY"
    nym_request.operation[VERKEY] = "HNjfjoeZ7WAHYDSzWcvzyvUABepctabD7QSxopM48fYz"
    txn = reqToTxn(nym_request)
    seq_no = 1
    txn_time = 1560241033
    append_txn_metadata(txn, seq_no, txn_time)
    with pytest.raises(InvalidClientRequest) as e:
        nym_handler._validate_new_nym(nym_request, nym_request.operation)
    e.match("Identifier with version 1")


def test_nym_validation_legacy_convention(nym_request: Request, nym_handler: NymHandler, doc, nym_request_factory):
    nym_request = nym_request_factory(doc, NYM_VERSION_CONVENTION)
    nym_request.operation[TARGET_NYM] = "X3XUxYQM2cfkSMzfMNma73"
    nym_request.operation[VERKEY] = "HNjfjoeZ7WAHYDSzWcvzyvUABepctabD7QSxopM48fYz"
    txn = reqToTxn(nym_request)
    seq_no = 1
    txn_time = 1560241033
    append_txn_metadata(txn, seq_no, txn_time)

    with mock.patch.object(nym_handler, "write_req_validator"):
        nym_handler._validate_new_nym(nym_request, nym_request.operation)


def test_nym_validation_self_certifying_x(nym_request: Request, nym_handler: NymHandler, doc, nym_request_factory):
    nym_request = nym_request_factory(doc, NYM_VERSION_SELF_CERT)
    nym_request.operation[TARGET_NYM] = "NOTSELFCERTIFYING"
    nym_request.operation[VERKEY] = "HNjfjoeZ7WAHYDSzWcvzyvUABepctabD7QSxopM48fYz"
    txn = reqToTxn(nym_request)
    seq_no = 1
    txn_time = 1560241033
    append_txn_metadata(txn, seq_no, txn_time)
    with pytest.raises(InvalidClientRequest) as e:
        nym_handler._validate_new_nym(nym_request, nym_request.operation)
    e.match("Identifier with version 2")


def test_nym_validation_self_certifying(nym_request: Request, nym_handler: NymHandler, doc, nym_request_factory):
    nym_request = nym_request_factory(doc, NYM_VERSION_SELF_CERT)
    nym_request.operation[TARGET_NYM] = "Dyqasan6xG5KsKdLufxCEf"
    nym_request.operation[VERKEY] = "HNjfjoeZ7WAHYDSzWcvzyvUABepctabD7QSxopM48fYz"
    txn = reqToTxn(nym_request)
    seq_no = 1
    txn_time = 1560241033
    append_txn_metadata(txn, seq_no, txn_time)

    with mock.patch.object(nym_handler, "write_req_validator"):
        nym_handler._validate_new_nym(nym_request, nym_request.operation)



import pytest
import json

from common.serializers.serialization import config_state_serializer
from indy_common.state import config
from indy_node.server.request_handlers.config_req_handlers.auth_rule.static_auth_rule_helper import StaticAuthRuleHelper
from plenum.common.constants import TRUSTEE, STEWARD, DATA
from plenum.common.exceptions import RequestRejectedException, \
    RequestNackedException

from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.authorize.auth_constraints import ROLE, CONSTRAINT_ID, ConstraintsEnum, SIG_COUNT, NEED_TO_BE_OWNER, \
    METADATA, OFF_LEDGER_SIGNATURE
from indy_common.constants import AUTH_ACTION, OLD_VALUE, NYM, ENDORSER, CONFIG_LEDGER_ID, CONSTRAINT

from plenum.test.helper import sdk_gen_request, sdk_sign_and_submit_req_obj, sdk_get_and_check_replies
from indy_node.test.auth_rule.helper import sdk_send_and_check_req_json, sdk_send_and_check_get_auth_rule_request
from indy_node.test.auth_rule.helper import (
    generate_constraint_entity, generate_constraint_list,
    sdk_send_and_check_auth_rule_request, generate_auth_rule_operation,
    generate_key,
    sdk_send_and_check_auth_rule_invalid_request
)


def test_auth_rule_transaction_for_edit(looper,
                                        sdk_wallet_trustee,
                                        sdk_pool_handle):
    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=EDIT_PREFIX)


def test_auth_rule_transaction(looper,
                               sdk_wallet_trustee,
                               sdk_pool_handle):
    sdk_send_and_check_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee
    )


def test_auth_rule_transaction_with_large_constraint(looper,
                                                     sdk_wallet_trustee,
                                                     sdk_pool_handle):
    constraint = generate_constraint_list(auth_constraints=[generate_constraint_entity(role=TRUSTEE),
                                                            generate_constraint_entity(role=STEWARD)])
    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         constraint=constraint)


def test_reject_with_unacceptable_role_in_constraint(looper,
                                                     sdk_wallet_trustee,
                                                     sdk_pool_handle):
    constraint = generate_constraint_entity()
    unacceptable_role = 'olololo'
    constraint[ROLE] = unacceptable_role
    with pytest.raises(RequestNackedException) as e:
        sdk_send_and_check_auth_rule_request(looper,
                                             sdk_pool_handle,
                                             sdk_wallet_trustee,
                                             constraint=constraint)
    e.match('InvalidClientRequest')
    e.match('client request invalid')
    e.match('Role {} is not acceptable'.format(unacceptable_role))


def test_reject_auth_rule_transaction(looper,
                                      sdk_wallet_steward,
                                      sdk_pool_handle):
    with pytest.raises(RequestRejectedException) as e:
        sdk_send_and_check_auth_rule_request(
            looper, sdk_pool_handle, sdk_wallet_steward
        )
    e.match('Not enough TRUSTEE signatures')


def test_reqnack_auth_rule_transaction_with_wrong_key(looper,
                                                      sdk_wallet_trustee,
                                                      sdk_pool_handle):
    with pytest.raises(RequestNackedException) as e:
        sdk_send_and_check_auth_rule_request(looper,
                                             sdk_pool_handle,
                                             sdk_wallet_trustee,
                                             auth_type="*")
    e.match('InvalidClientRequest')
    e.match("client request invalid")
    e.match("is not found in authorization map")


def test_reqnack_auth_rule_edit_transaction_with_wrong_format(looper,
                                                              sdk_wallet_trustee,
                                                              sdk_pool_handle):
    op = generate_auth_rule_operation(auth_action=EDIT_PREFIX)
    op.pop(OLD_VALUE)

    req_obj = sdk_gen_request(op, identifier=sdk_wallet_trustee[1])
    req_json = json.dumps(req_obj.as_dict)
    with pytest.raises(RequestNackedException) as e:
        sdk_send_and_check_req_json(
            looper, sdk_pool_handle, sdk_wallet_trustee, req_json,
        )
    e.match('InvalidClientRequest')
    e.match("client request invalid")
    e.match("Transaction for change authentication "
            "rule for {}={} must contain field {}".
            format(AUTH_ACTION, EDIT_PREFIX, OLD_VALUE))


def test_reqnack_auth_rule_add_transaction_with_wrong_format(looper,
                                                             sdk_wallet_trustee,
                                                             sdk_pool_handle):
    with pytest.raises(RequestNackedException) as e:
        sdk_send_and_check_auth_rule_invalid_request(
            looper,
            sdk_pool_handle,
            sdk_wallet_trustee,
            **generate_key(old_value="*")
        )
    e.match('InvalidClientRequest')
    e.match("client request invalid")
    e.match("Transaction for change authentication "
            "rule for {}={} must not contain field {}".
            format(AUTH_ACTION, ADD_PREFIX, OLD_VALUE))


@pytest.mark.parametrize("off_ledger_signature", [True, False])
def test_auth_rule_state_format(
    looper, sdk_pool_handle, sdk_wallet_trustee, txnPoolNodeSet, off_ledger_signature
):
    auth_action = ADD_PREFIX
    auth_type = NYM
    field = ROLE
    new_value = ENDORSER
    constraint = {CONSTRAINT_ID: ConstraintsEnum.ROLE_CONSTRAINT_ID,
                  ROLE: "*",
                  SIG_COUNT: 1,
                  NEED_TO_BE_OWNER: False,
                  METADATA: {}}
    if off_ledger_signature:
        constraint[OFF_LEDGER_SIGNATURE] = off_ledger_signature
    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=auth_action,
                                         auth_type=auth_type,
                                         field=field,
                                         new_value=new_value,
                                         constraint=constraint)
    state = txnPoolNodeSet[0].db_manager.get_database(CONFIG_LEDGER_ID).state
    key = generate_key(auth_action, auth_type, field, new_value)
    path = config.make_state_path_for_auth_rule(StaticAuthRuleHelper.get_auth_key(key))
    state_constraint = config_state_serializer.deserialize(state.get(path))
    assert state_constraint == constraint

    _, before_resp = sdk_send_and_check_get_auth_rule_request(looper,
                                                              sdk_pool_handle,
                                                              sdk_wallet_trustee,
                                                              auth_type=auth_type,
                                                              auth_action=auth_action,
                                                              field=field,
                                                              new_value=new_value
                                                              )[0]

    for rule in before_resp["result"][DATA]:
        if rule[CONSTRAINT][CONSTRAINT_ID] == 'ROLE' and off_ledger_signature:
            assert OFF_LEDGER_SIGNATURE in rule[CONSTRAINT]



import csv
import io
import os
import datetime
import functools
from typing import Type, Union, Tuple
from enum import Enum, unique
from dateutil.parser import parse as parse_dt

# TODO
#  - move csv serializer to separate module


class CsvSerializer:
    _items = []

    def __iter__(self):
        for item in self._items:
            yield getattr(self, item)

    def pack(self, delimiter: str = '\t'):
        with io.StringIO() as fs:
            csv.writer(fs, delimiter=delimiter).writerow(self)
            return fs.getvalue()

    @classmethod
    def unpack(cls, row: str, *args, delimiter: str = '\t', **kwargs):
        reader = csv.reader([row], delimiter=delimiter)
        return cls(*next(reader), *args, **kwargs)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return str(self.__dict__)


class ActionLogData(CsvSerializer):
    _items = ['when']

    def __init__(self, when: Union[datetime.datetime, str]):
        if isinstance(when, str):
            when = parse_dt(when)
        if not isinstance(when, datetime.datetime):
            raise TypeError(
                "'when' should be 'datetime' or 'str', got: {}"
                .format(type(when))
            )
        self.when = when


@unique
class ActionLogEvents(Enum):
    scheduled = 1
    started = 2
    succeeded = 3
    failed = 4
    cancelled = 5

    def __str__(self):
        return self.name


# Note. datetime class is referred not directly (through datetime module)
# since it makes possible to mock it in tests using subclasses
class ActionLogEvent(CsvSerializer):
    def __init__(
            self,
            ts: Union[datetime.datetime, str],
            ev_type: Enum,
            data: Type[Union[CsvSerializer, str]],
            *args: Tuple[str],
            data_class: Type[Union[CsvSerializer, None]] = None,
            types: Type[Enum] = ActionLogEvents
    ):
        if ts:
            if isinstance(ts, str):
                ts = parse_dt(ts)
            if not isinstance(ts, datetime.datetime):
                raise TypeError(
                    "'ts' should be 'datetime' or None, got: {}"
                    .format(type(ts))
                )

        if isinstance(ev_type, str):
            try:
                ev_type = types[ev_type]
            except KeyError:
                raise ValueError("Unknown event {}"
                                 .format(ev_type))

        if not isinstance(ev_type, Enum):
            raise TypeError(
                "'ev_type' should be 'Enum', got: {}"
                .format(type(ts))
            )

        if ev_type not in types:
            raise ValueError("Unknown event {}".format(ev_type))

        data = data_class(data, *args) if data_class else data
        if not isinstance(data, CsvSerializer):
            raise TypeError(
                "'data' should be 'CsvSerializer', got: {}"
                .format(type(data))
            )

        self.ts = ts if ts else datetime.datetime.utcnow()
        self.ev_type = ev_type
        self.data = data

        self._data_items_prefix = '_data_'
        self._items = (
            ['ts', 'ev_type'] + [(self._data_items_prefix + i) for i in self.data._items]
        )

    def __getattr__(self, name):
        try:
            _name = name.split(self._data_items_prefix)[1]
        except IndexError:
            raise AttributeError
        else:
            return getattr(self.data, _name)


class ActionLog:
    """
    Append-only event log of action event
    """

    def __init__(
            self,
            file_path: str,
            data_class: Type[CsvSerializer] = ActionLogData,
            event_types: Type[Enum] = ActionLogEvents,
            delimiter: str = '\t'
    ):
        self._delimiter = delimiter
        self._file_path = file_path
        self._items = []
        self._data_class = data_class
        self._event_types = event_types
        self._load()

        for ev_type in self._event_types:
            setattr(self, "append_{}".format(ev_type.name),
                    functools.partial(self._append, ev_type))

    @property
    def file_path(self) -> str:
        return self._file_path

    @property
    def delimiter(self) -> str:
        return self._delimiter

    @property
    def last_event(self) -> ActionLogEvent:
        return self._items[-1] if self._items else None

    @property
    def event_types(self) -> Enum:
        return self._event_types

    def _load(self):
        if os.path.exists(self._file_path):
            with open(self._file_path, mode='r', newline='') as f:
                for line in f:
                    self._items.append(ActionLogEvent.unpack(
                        line,
                        delimiter=self._delimiter,
                        data_class=self._data_class,
                        types=self._event_types
                    ))

    def _append(self, ev_type: Enum, data: ActionLogData) -> None:
        """
        Appends event to log
        Be careful it opens file every time!
        """
        event = ActionLogEvent(None, ev_type, data, types=self._event_types)
        with open(self._file_path, mode='a+', newline='') as f:
            f.write(event.pack(delimiter=self._delimiter))
        self._items.append(event)

    def __iter__(self):
        for item in self._items:
            yield item

    def __len__(self):
        return len(self._items)



import inspect

import typing

"""
Thanks to Ilya Peterov for the base code from
https://github.com/ipeterov/strict_types.
"""

if 'defaultShouldCheck' not in globals():
    defaultShouldCheck = False


class strict_types:

    def __init__(self, shouldCheck=None):
        if shouldCheck is not None:
            self.shouldCheck = shouldCheck
        else:
            self.shouldCheck = defaultShouldCheck

    def is_complex_type(self, type_):
        complex_types = [type(typing.Union), type(typing.Tuple)]
        return any(isinstance(type_, complex_type)
                   for complex_type in complex_types)

    def is_subtype(self, type_a, type_b):
        # This wouldn't work for nested Types (from typing package)
        # like Union[Tuple[...]] but since there is no such types in the
        # project (at least for now) this simple implementation is okay

        if self.is_complex_type(type_b):
            type_b = tuple(
                getattr(type_b, '__args__', None) or getattr(type_b, '__union_set_params__', None)
            )

        if self.is_complex_type(type_a):
            return type_a is type_b
        return issubclass(type_a, type_b)

    def __call__(self, function):

        if not self.shouldCheck:
            return function

        type_hints = typing.get_type_hints(function)

        def precheck(*args, **kwargs):

            all_args = kwargs.copy()
            all_args.update(dict(zip(function.__code__.co_varnames, args)))
            runtime_args = ((n, type(v)) for n, v in all_args.items())

            for arg_name, arg_type in runtime_args:
                if arg_name not in type_hints:
                    continue
                if not self.is_subtype(arg_type, type_hints[arg_name]):
                    raise TypeError('In {} type of {} is {} and not {}'.
                                    format(function.__qualname__,
                                           arg_name,
                                           arg_type,
                                           type_hints[arg_name]))

        def postcheck(result):
            if 'return' in type_hints:
                if not self.is_subtype(type(result), type_hints['return']):
                    raise TypeError('Type of result is {} and not {}'.
                                    format(type(result), type_hints['return']))
            return result

        if inspect.iscoroutinefunction(function):
            async def type_checker(*args, **kwargs):
                precheck(*args, **kwargs)
                result = await function(*args, **kwargs)
                return postcheck(result)
        else:
            def type_checker(*args, **kwargs):
                precheck(*args, **kwargs)
                result = function(*args, **kwargs)
                return postcheck(result)

        return type_checker


def decClassMethods(decorator):
    def decClass(cls):
        for name, m in inspect.getmembers(cls, inspect.isfunction):
            setattr(cls, name, decorator(m))
        return cls
    return decClass



import os
import shutil

from stp_core.common.log import getlogger

logger = getlogger()

ENV_FILE_PATH = "/etc/indy/indy.env"


def update_indy_env(key, val):
    key_found = False

    ENV_FILE_PATH_TMP = ENV_FILE_PATH + "_tmp"
    logger.info("Make a copy '{}' of an original env file '{}'."
                .format(ENV_FILE_PATH_TMP, ENV_FILE_PATH))
    # Make a copy to inherit r/w/own flags.
    shutil.copy(ENV_FILE_PATH, ENV_FILE_PATH_TMP)

    new_line = "{}={}\n".format(key, val)

    with open(ENV_FILE_PATH, "r") as old_env_file:
        with open(ENV_FILE_PATH_TMP, "w") as new_env_file:
            for line in old_env_file.readlines():
                _key = line.split('=')[0].strip()
                if key == _key:
                    key_found = True
                    line = new_line
                logger.info("Write '{}' line to '{}'.".format(line.strip(), ENV_FILE_PATH_TMP))
                new_env_file.write(line)

    if not key_found:
        logger.info("Key '{}' is not found in '{}', append line."
                    .format(key, ENV_FILE_PATH))
        with open(ENV_FILE_PATH, "a") as env_file:
            env_file.write("\n" + new_line)
        logger.info("Remove unnecessary '{}'.".format(ENV_FILE_PATH_TMP))
        os.remove(ENV_FILE_PATH_TMP)
    else:
        logger.info("Rename '{}' to '{}'.".format(ENV_FILE_PATH_TMP, ENV_FILE_PATH))
        shutil.move(ENV_FILE_PATH_TMP, ENV_FILE_PATH)



import os

import importlib
import pytest
from plenum.test.recorder.helper import reload_modules_for_recorder, _reload_module
from stp_core.common.log import getlogger, Logger
import tempfile

from indy_node.test.catchup.conftest import some_transactions_done

whitelist = ['cannot find remote with name']


def find_ns_script_dir():
    path = os.path.dirname(__file__)
    path = os.path.join(path, '../../..', 'tools', 'diagnostics')
    path = os.path.abspath(path)
    return path


def load_script(temp_file, script_name):
    spec = importlib.util.spec_from_file_location(script_name, temp_file)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod


def run_script(script_name, args):
    cmd_path = os.path.join(find_ns_script_dir(), script_name)
    with open(cmd_path, 'r+b') as r_file:
        file_content = r_file.read()
        with tempfile.NamedTemporaryFile('w+b', suffix=".py", delete=False) as w_file:
            w_file.write(file_content)
            w_file.flush()

            mod = load_script(w_file.name, script_name)

            args = mod.parse_args(argv=args)
            try:
                rtn = mod.main(args)
            except SystemExit:
                return -1
            return rtn


def test_nscapture_unit_tests():
    cmd_path = os.path.join(find_ns_script_dir(), 'nscapture')
    with open(cmd_path, 'r+b') as r_file:
        file_content = r_file.read()
        with tempfile.NamedTemporaryFile('w+b', suffix=".py") as w_file:
            w_file.write(file_content)
            w_file.flush()

            mod = load_script(w_file.name, 'nscapture')

            rtn = mod.test(None, mod)
            assert rtn == 0, "The internal unit test did not pass"


@pytest.fixture(autouse=True)
def setup_logging(tdir, tconf):
    Logger().apply_config(tconf)
    log_file_name = os.path.join(tdir, 'var', 'log', 'indy', 'sandbox',
                                 'Alpha' + ".log")
    Logger().enableFileLogging(log_file_name)
    logger = getlogger()
    logger.setLevel(tconf.logLevel)


@pytest.fixture(scope="module")
def tconf(tconf):
    tconf.STACK_COMPANION = 1
    reload_modules_for_recorder(tconf)
    import indy_node.server.node
    import indy_node.test.helper
    _reload_module(indy_node.server.node)
    _reload_module(indy_node.test.helper)
    return tconf


@pytest.fixture(scope="session", autouse=True)
def warncheck(warnfilters):
    pass


@pytest.mark.skip(reason="Need to fix nsreplay due to OrderingService integration")
def test_end_to_end_replay(looper,
                           tdir,
                           tdirWithPoolTxns,
                           tdirWithDomainTxns,
                           nodeSet,
                           sdk_pool_handle,
                           sdk_wallet_endorser
                           ):
    some_transactions_done(looper, nodeSet, sdk_pool_handle,
                           sdk_wallet_endorser)

    for node in nodeSet:
        node.cleanupOnStopping = False
        node.stop()

    root_dir = tdir
    # root_dir = '/home/devin/temp/B6/test/0'

    #  Capture the original
    rtn = run_script('nscapture', ['-r', root_dir, '-n', 'Alpha', '-o', tdir])
    assert rtn == 0, "Capture did not run successfully"

    recording = list(filter(lambda x: x.startswith('Alpha'), os.listdir(tdir)))
    assert len(recording) == 1
    recording = recording.pop()

    #  Replay recording
    replay_path = os.path.join(tdir, 'replayed', '0')
    rtn = run_script('nsreplay', ['-o', replay_path, os.path.join(tdir, recording)])
    assert rtn == 0, "Replay did not run successfully"

    #  Capture the replay
    rtn = run_script('nscapture', ['-r', replay_path, '-n', 'Alpha', '-o', tdir])
    assert rtn == 0, "Capture did not run successfully"

    captures_archives = list(filter(lambda x: x.startswith('Alpha'), os.listdir(tdir)))
    assert len(captures_archives) == 2

    args = []
    for archive in captures_archives:
        args.append(os.path.join(tdir, archive))
    rtn = run_script('nsdiff', args)
    assert rtn == 0, "Recording and replay were not identical"
    pass



from indy_common.authorize.auth_request_validator import WriteRequestValidator

from indy_common.constants import RICH_SCHEMA_ENCODING, RS_ENC_INPUT, RS_ENC_OUTPUT, RS_ENC_ALGORITHM, RS_ENC_TEST_VECS, \
    RS_ENC_ID, RS_ENC_TYPE, RS_ENC_ALG_DESC, RS_ENC_ALG_DOC, RS_ENC_ALG_IMPL, RS_CONTENT

from indy_node.server.request_handlers.domain_req_handlers.rich_schema.abstract_rich_schema_object_handler import \
    AbstractRichSchemaObjectHandler
from plenum.common.exceptions import InvalidClientRequest

from plenum.server.database_manager import DatabaseManager
from stp_core.common.log import getlogger

logger = getlogger()


class RichSchemaEncodingHandler(AbstractRichSchemaObjectHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(RICH_SCHEMA_ENCODING, database_manager, write_req_validator)

    def is_json_ld_content(self):
        return False

    def do_static_validation_content(self, content_as_dict, request):
        # 1. check for top level fields
        missing_fields = []
        for field in [RS_ENC_INPUT, RS_ENC_OUTPUT, RS_ENC_ALGORITHM, RS_ENC_TEST_VECS]:
            if not content_as_dict.get(field):
                missing_fields.append("'{}'".format(field))
        if missing_fields:
            missing_fields_str = ", ".join(missing_fields)
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "{} must be set in '{}'".format(missing_fields_str, RS_CONTENT))

        # 2. check for input-output fields
        for io_field in [RS_ENC_INPUT, RS_ENC_OUTPUT]:
            missing_io_fields = []
            for field in [RS_ENC_ID, RS_ENC_TYPE]:
                if not content_as_dict[io_field].get(field):
                    missing_io_fields.append("'{}'".format(field))
            if missing_io_fields:
                missing_io_fields_str = " and ".join(missing_io_fields)
                raise InvalidClientRequest(request.identifier, request.reqId,
                                           "{} must be set in '{}'".format(missing_io_fields_str, io_field))

        # 3. check for algorithm fields
        missing_alg_fields = []
        for field in [RS_ENC_ALG_DESC, RS_ENC_ALG_DOC, RS_ENC_ALG_IMPL]:
            if not content_as_dict[RS_ENC_ALGORITHM].get(field):
                missing_alg_fields.append("'{}'".format(field))
        if missing_alg_fields:
            missing_alg_fields_str = ", ".join(missing_alg_fields)
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "{} must be set in '{}'".format(missing_alg_fields_str, RS_ENC_ALGORITHM))

    def do_dynamic_validation_content(self, request):
        pass



import pytest as pytest

from indy_node.server.request_handlers.config_req_handlers.txn_author_agreement_handler_v1 import \
    TxnAuthorAgreementHandlerV1
from plenum.common.constants import TXN_AUTHOR_AGREEMENT_TEXT, TXN_AUTHOR_AGREEMENT_VERSION, DOMAIN_LEDGER_ID
from plenum.common.txn_util import reqToTxn, get_payload_data, append_txn_metadata
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.static_taa_helper import StaticTAAHelper
from plenum.test.testing_utils import FakeSomething
from state.state import State


@pytest.fixture(scope="function")
def txn_author_agreement_handler(tconf, domain_state, write_auth_req_validator):
    data_manager = DatabaseManager()
    handler = TxnAuthorAgreementHandlerV1(data_manager, write_auth_req_validator)
    state = State()
    state.txn_list = {}
    state.get = lambda key, isCommitted=False: state.txn_list.get(key, None)
    state.set = lambda key, value, isCommitted=False: state.txn_list.update({key: value})
    data_manager.register_new_database(handler.ledger_id,
                                       FakeSomething(),
                                       state)
    data_manager.register_new_database(DOMAIN_LEDGER_ID,
                                       FakeSomething(),
                                       domain_state)
    return handler


def test_update_state(txn_author_agreement_handler, taa_request):
    seq_no = 1
    txn_time = 1560241033
    txn_id = "id"
    txn = reqToTxn(taa_request)
    payload = get_payload_data(txn)
    text = payload[TXN_AUTHOR_AGREEMENT_TEXT]
    version = payload[TXN_AUTHOR_AGREEMENT_VERSION]
    digest = StaticTAAHelper.taa_digest(text, version)
    append_txn_metadata(txn, seq_no, txn_time, txn_id)

    state_value = {TXN_AUTHOR_AGREEMENT_TEXT: text,
                   TXN_AUTHOR_AGREEMENT_VERSION: version}

    txn_author_agreement_handler.update_state(txn, None, taa_request)

    assert txn_author_agreement_handler.get_from_state(
        StaticTAAHelper.state_path_taa_digest(digest)) == (state_value, seq_no, txn_time)
    assert txn_author_agreement_handler.state.get(
        StaticTAAHelper.state_path_taa_latest()) == digest
    assert txn_author_agreement_handler.state.get(
        StaticTAAHelper.state_path_taa_version(version)) == digest



from typing import NamedTuple, List, Optional

from indy_common.authorize.auth_actions import AbstractAuthAction, AuthActionAdd
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.authorize.authorizer import AbstractAuthorizer
from indy_common.test.constants import IDENTIFIERS
from indy_common.types import Request
from plenum.common.constants import TYPE
from plenum.test.helper import randomOperation

PLUGIN_FIELD = "new_field"

Action = NamedTuple('Action',
                    [("author", str), ("endorser", Optional[str]), ("sigs", dict),
                     ("is_owner", bool), ("amount", int), ("extra_sigs", bool)])


class PluginAuthorizer(AbstractAuthorizer):

    def authorize(self,
                  request: Request,
                  auth_constraint: AuthConstraint,
                  auth_action: AbstractAuthAction):
        if not auth_constraint.metadata and PLUGIN_FIELD in request.operation:
            return False, "plugin field must be absent"
        if not auth_constraint.metadata:
            return True, ""
        if PLUGIN_FIELD not in auth_constraint.metadata:
            return True, ""
        if PLUGIN_FIELD not in request.operation:
            return False, "missing required plugin field"

        required_amount = auth_constraint.metadata[PLUGIN_FIELD]
        req_amount = request.operation[PLUGIN_FIELD]

        if req_amount != required_amount:
            return False, "not enough amount in plugin field"

        return True, ""


def set_auth_constraint(validator, auth_constraint):
    validator.auth_cons_strategy.get_auth_constraint = lambda a: auth_constraint


def build_req_and_action(action: Action):
    sig = None
    sigs = None
    identifier = IDENTIFIERS[action.author][0]
    endorser_did = None

    # if there is only 1 sig from the Author - use `signature` instead of `signatures`
    if len(action.sigs) == 1 and next(iter(action.sigs.values())) == 1 and next(
            iter(action.sigs.keys())) == action.author:
        sig = 'signature'
    else:
        sigs = {IDENTIFIERS[role][i]: 'signature' for role, sig_count in action.sigs.items() for i in
                range(sig_count)}

    if action.endorser is not None:
        endorser_did = IDENTIFIERS[action.endorser][0]

    operation = randomOperation()
    if action.amount is not None:
        operation[PLUGIN_FIELD] = action.amount

    req = Request(identifier=identifier,
                  operation=operation,
                  signature=sig,
                  signatures=sigs,
                  endorser=endorser_did)
    action = AuthActionAdd(txn_type=req.operation[TYPE],
                           field='some_field',
                           value='new_value',
                           is_owner=action.is_owner)

    return req, [action]


def validate(auth_constraint,
             valid_actions: List[Action],
             author, endorser, all_signatures, is_owner, amount,
             write_auth_req_validator,
             write_request_validation):
    set_auth_constraint(write_auth_req_validator,
                        auth_constraint)

    for signatures in all_signatures:
        next_action = Action(author=author, endorser=endorser, sigs=signatures,
                             is_owner=is_owner, amount=amount, extra_sigs=False)
        expected = is_expected(next_action, valid_actions)
        result = write_request_validation(*build_req_and_action(next_action))
        assert result == expected, \
            "Expected {} but result is {} for case {} and valid set {}".format(expected, result, next_action,
                                                                               valid_actions)


def is_expected(next_action: Action, valid_actions: List[Action]):
    for valid_action in valid_actions:
        if (next_action.author, next_action.endorser, next_action.is_owner, next_action.amount) != \
                (valid_action.author, valid_action.endorser, valid_action.is_owner, valid_action.amount):
            continue
        if not valid_action.extra_sigs and next_action.sigs != valid_action.sigs:
            continue
        if valid_action.extra_sigs and not (valid_action.sigs.items() <= next_action.sigs.items()):
            continue
        return True
    return False



#!/usr/bin/python3.5
import os
import shutil

import subprocess
from common.serializers.compact_serializer import CompactSerializer
from common.serializers.json_serializer import JsonSerializer
from common.serializers.mapping_serializer import MappingSerializer
from ledger.compact_merkle_tree import CompactMerkleTree
from ledger.ledger import Ledger
from plenum.persistence.leveldb_hash_store import LevelDbHashStore
from storage import store_utils
from storage.chunked_file_store import ChunkedFileStore

from indy_common.config_util import getConfig
from indy_common.txn_util import getTxnOrderedFields
from stp_core.common.log import getlogger

config = getConfig()
logger = getlogger()


def _migrate_ledger(data_directory,
                    old_ledger_file, new_ledger_file,
                    serializer: MappingSerializer = None):
    """
    Test for the directory, open old and new ledger, migrate data, rename directories
    """

    # we should have ChunkedFileStorage implementation of the Ledger
    if not os.path.isdir(os.path.join(data_directory, old_ledger_file)):
        msg = 'Could not find directory {} for migration.'.format(
            old_ledger_file)
        logger.error(msg)
        raise Exception(msg)

    # open the old ledger using the specified serializer
    old_ledger_file_backup = old_ledger_file + "_new"
    old_txn_log_store = ChunkedFileStore(data_directory,
                                         old_ledger_file_backup,
                                         isLineNoKey=True,
                                         storeContentHash=False)
    old_ledger = Ledger(CompactMerkleTree(),
                        dataDir=data_directory,
                        txn_serializer=serializer,
                        hash_serializer=serializer,
                        fileName=old_ledger_file_backup,
                        transactionLogStore=old_txn_log_store)

    # open the new ledger with new serialization
    new_ledger = Ledger(CompactMerkleTree(),
                        dataDir=data_directory,
                        fileName=new_ledger_file)
    logger.info("new size for {}: {}".format(
        old_ledger_file_backup, str(new_ledger.size)))

    # add all txns into the old ledger
    for _, txn in new_ledger.getAllTxn():
        old_ledger.add(txn)
    logger.info("old size for {}: {}".format(
        new_ledger_file, str(old_ledger.size)))

    old_ledger.stop()
    new_ledger.stop()

    # now that everything succeeded, remove the new files and move the old
    # files into place
    shutil.rmtree(
        os.path.join(data_directory, new_ledger_file))
    os.rename(
        os.path.join(data_directory, old_ledger_file_backup),
        os.path.join(data_directory, old_ledger_file))


def _open_old_ledger(data_directory, old_ledger_file,
                     hash_store_name, serializer):
    # open old Ledger with leveldb hash store (to re-init it)
    old_txn_log_store = ChunkedFileStore(data_directory,
                                         old_ledger_file,
                                         isLineNoKey=True,
                                         storeContentHash=False)
    old_ledger = Ledger(CompactMerkleTree(
        hashStore=LevelDbHashStore(
            dataDir=data_directory,
            fileNamePrefix=hash_store_name)),
        dataDir=data_directory,
        txn_serializer=serializer,
        hash_serializer=serializer,
        fileName=old_ledger_file,
        transactionLogStore=old_txn_log_store)

    old_ledger.stop()


def migrate_all_hash_stores(node_data_directory):
    # the new hash store (merkle tree) will be recovered from the new transaction log after re-start
    # just delete the current hash store
    new_merkle_nodes = os.path.join(node_data_directory, '_merkleNodes')
    new_merkle_leaves = os.path.join(node_data_directory, '_merkleLeaves')
    new_merkle_nodes_bin = os.path.join(
        node_data_directory, '_merkleNodes.bin')
    new_merkle_leaves_bin = os.path.join(
        node_data_directory, '_merkleLeaves.bin')
    new_merkle_nodes_config_bin = os.path.join(
        node_data_directory, 'config_merkleNodes.bin')
    new_merkle_leaves_config_bin = os.path.join(
        node_data_directory, 'config_merkleLeaves.bin')

    if os.path.exists(new_merkle_nodes):
        shutil.rmtree(new_merkle_nodes)
    if os.path.exists(new_merkle_leaves):
        shutil.rmtree(new_merkle_leaves)
    if os.path.exists(new_merkle_nodes_bin):
        os.remove(new_merkle_nodes_bin)
    if os.path.exists(new_merkle_leaves_bin):
        os.remove(new_merkle_leaves_bin)
    if os.path.exists(new_merkle_nodes_config_bin):
        os.remove(new_merkle_nodes_config_bin)
    if os.path.exists(new_merkle_leaves_config_bin):
        os.remove(new_merkle_leaves_config_bin)

    # open new Ledgers
    fields = getTxnOrderedFields()
    _open_old_ledger(node_data_directory, config.poolTransactionsFile,
                     'pool', serializer=JsonSerializer())
    _open_old_ledger(node_data_directory, config.domainTransactionsFile,
                     'domain', serializer=CompactSerializer(fields=fields))
    _open_old_ledger(node_data_directory, config.configTransactionsFile,
                     'config', serializer=JsonSerializer())


def migrate_all_ledgers_for_node(node_data_directory):
    # using default ledger names
    _migrate_ledger(node_data_directory,
                    config.poolTransactionsFile, config.poolTransactionsFile,
                    serializer=JsonSerializer())
    _migrate_ledger(
        node_data_directory,
        config.configTransactionsFile,
        config.configTransactionsFile,
        serializer=JsonSerializer())

    # domain ledger uses custom CompactSerializer and old file name
    fields = getTxnOrderedFields()
    _migrate_ledger(node_data_directory,
                    config.domainTransactionsFile.replace(
                        'domain_', ''), config.domainTransactionsFile,
                    serializer=CompactSerializer(fields=fields))


def migrate_all_states(node_data_directory):
    # the states will be recovered from the ledger during the start-up.
    # just delete the current ones
    shutil.rmtree(
        os.path.join(node_data_directory, 'pool_state'))
    shutil.rmtree(
        os.path.join(node_data_directory, 'domain_state'))
    shutil.rmtree(
        os.path.join(node_data_directory, 'config_state'))


def migrate_genesis_txn(base_dir):
    for suffix in ('sandbox', 'live', 'local'):
        old_domain_genesis = os.path.join(
            base_dir, 'transactions_{}'.format(suffix))
        old_pool_genesis = os.path.join(
            base_dir, 'pool_transactions_{}'.format(suffix))

        new_domain_genesis = os.path.join(
            base_dir, 'domain_transactions_{}_genesis'.format(suffix))
        new_pool_genesis = os.path.join(
            base_dir, 'pool_transactions_{}_genesis'.format(suffix))

        if os.path.exists(old_domain_genesis):
            os.remove(old_domain_genesis)
        if os.path.exists(old_pool_genesis):
            os.remove(old_pool_genesis)

        if os.path.exists(new_domain_genesis):
            old_ser = CompactSerializer(getTxnOrderedFields())
            new_ser = JsonSerializer()
            with open(new_domain_genesis, 'r') as f1:
                with open(old_domain_genesis, 'w') as f2:
                    for line in store_utils.cleanLines(f1):
                        txn = new_ser.deserialize(line)
                        txn = old_ser.serialize(txn)
                        f2.write(txn)
            os.remove(new_domain_genesis)
        if os.path.exists(new_pool_genesis):
            os.rename(new_pool_genesis, old_domain_genesis)


def migrate_all():
    base_dir = config.baseDir
    nodes_data_dir = os.path.join(base_dir, config.nodeDataDir)
    if not os.path.exists(nodes_data_dir):
        # TODO: find a better way
        base_dir = '/home/sovrin/.sovrin'
        nodes_data_dir = os.path.join(base_dir, config.nodeDataDir)
    if not os.path.exists(nodes_data_dir):
        msg = 'Can not find the directory with the ledger: {}'.format(
            nodes_data_dir)
        logger.error(msg)
        raise Exception(msg)

    for node_dir in os.listdir(nodes_data_dir):
        node_data_dir = os.path.join(nodes_data_dir, node_dir)
        migrate_all_ledgers_for_node(node_data_dir)
        migrate_all_hash_stores(node_data_dir)
        migrate_all_states(node_data_dir)

    migrate_genesis_txn(base_dir)
    subprocess.run(['chown', '-R', 'sovrin:sovrin', '/home/sovrin/.sovrin'])


migrate_all()



import pytest
import time
from indy_common.constants import REVOC_REG_DEF_ID, ISSUED, REVOKED, \
    PREV_ACCUM, ACCUM, TXN_TYPE, CRED_DEF_ID, VALUE, TAG
from indy_common.types import Request
from indy_common.state import domain
from plenum.common.types import f
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.util import randomString


def test_validation_with_prev_accum_but_empty_ledger(
        build_txn_for_revoc_def_entry_by_default,
        create_node_and_not_start):
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_default
    del req_entry['operation'][VALUE][ISSUED]
    del req_entry['operation'][VALUE][REVOKED]
    with pytest.raises(InvalidClientRequest, match="but there is no previous"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)


def test_validation_with_right_accums_but_empty_indices(
        build_txn_for_revoc_def_entry_by_default,
        create_node_and_not_start):
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_default
    node.write_manager.apply_request(Request(**req_entry), int(time.time()))
    req_entry['operation'][VALUE][PREV_ACCUM] = req_entry['operation'][VALUE][ACCUM]
    req_entry['operation'][VALUE][ACCUM] = randomString(10)
    del req_entry['operation'][VALUE][ISSUED]
    del req_entry['operation'][VALUE][REVOKED]
    with pytest.raises(InvalidClientRequest, match="lists are empty"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)


def test_validation_with_unexpected_accum(
        build_txn_for_revoc_def_entry_by_default,
        create_node_and_not_start):
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_default
    node.write_manager.apply_request(Request(**req_entry), int(time.time()))
    with pytest.raises(InvalidClientRequest, match="must be equal to the last accumulator value"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)


def test_validation_with_same_revoked_by_default(
        build_txn_for_revoc_def_entry_by_default,
        create_node_and_not_start):
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_default
    req_entry['operation'][VALUE][REVOKED] = [1, 2]
    node.write_manager.apply_request(Request(**req_entry), int(time.time()))
    req_entry['operation'][VALUE][PREV_ACCUM] = req_entry['operation'][VALUE][ACCUM]
    req_entry['operation'][VALUE][ACCUM] = randomString(10)
    with pytest.raises(InvalidClientRequest, match="are already revoked in current state"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)


def test_validation_with_issued_no_revoked_before_by_default(
        build_txn_for_revoc_def_entry_by_default,
        create_node_and_not_start):
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_default
    req_entry['operation'][VALUE][REVOKED] = [1, 2]
    node.write_manager.apply_request(Request(**req_entry), int(time.time()))
    req_entry['operation'][VALUE][ISSUED] = [3, 4]
    req_entry['operation'][VALUE][REVOKED] = []
    req_entry['operation'][VALUE][PREV_ACCUM] = req_entry['operation'][VALUE][ACCUM]
    req_entry['operation'][VALUE][ACCUM] = randomString(10)
    with pytest.raises(InvalidClientRequest, match="are not present in the current revoked list"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)


def test_validation_with_same_issued_by_demand(
        build_txn_for_revoc_def_entry_by_demand,
        create_node_and_not_start):
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_demand
    req_entry['operation'][VALUE][ISSUED] = [1, 2]
    node.write_manager.apply_request(Request(**req_entry), int(time.time()))
    req_entry['operation'][VALUE][PREV_ACCUM] = req_entry['operation'][VALUE][ACCUM]
    req_entry['operation'][VALUE][ACCUM] = randomString(10)
    req_entry['operation'][VALUE][ISSUED] = [1, 2]
    with pytest.raises(InvalidClientRequest, match="are already issued in current state"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)


def test_validation_with_revoked_no_issued_before_by_demand(
        build_txn_for_revoc_def_entry_by_demand,
        create_node_and_not_start):
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_demand
    req_entry['operation'][VALUE][ISSUED] = [1, 2]
    node.write_manager.apply_request(Request(**req_entry), int(time.time()))
    req_entry['operation'][VALUE][REVOKED] = [3, 4]
    req_entry['operation'][VALUE][PREV_ACCUM] = req_entry['operation'][VALUE][ACCUM]
    req_entry['operation'][VALUE][ACCUM] = randomString(10)
    with pytest.raises(InvalidClientRequest, match="are not present in the current issued list"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)


def test_validation_if_issued_revoked_has_same_index(
        build_txn_for_revoc_def_entry_by_default,
        create_node_and_not_start):
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_default
    req_entry['operation'][VALUE][REVOKED] = [1, 2]
    req_entry['operation'][VALUE][ISSUED] = [1, 2]
    with pytest.raises(InvalidClientRequest, match="Can not have an index in both 'issued' and 'revoked' lists"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)


def test_validation_if_revoc_def_does_not_exist(
        build_txn_for_revoc_def_entry_by_default,
        create_node_and_not_start):
    path = ":".join([f.IDENTIFIER.nm,
                     domain.MARKER_REVOC_DEF,
                     CRED_DEF_ID,
                     TXN_TYPE,
                     TAG])
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_default
    req_entry['operation'][REVOC_REG_DEF_ID] = path
    with pytest.raises(InvalidClientRequest, match="There is no any REVOC_REG_DEF"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)


def test_validation_with_equal_accums_but_not_empty_indices(
        build_txn_for_revoc_def_entry_by_default,
        create_node_and_not_start):
    node = create_node_and_not_start
    req_entry = build_txn_for_revoc_def_entry_by_default
    node.write_manager.apply_request(Request(**req_entry), int(time.time()))
    req_entry['operation'][VALUE][PREV_ACCUM] = req_entry['operation'][VALUE][ACCUM]
    req_entry['operation'][VALUE][REVOKED] = [100, 200]
    with pytest.raises(InvalidClientRequest, match="Got equal accum and prev_accum"):
        node.write_manager.dynamic_validation(Request(**req_entry), 0)



from plenum.common.constants import RAW, ENC, HASH


class StateValue:
    def __init__(self, root_hash=None, value=None, seq_no=None, update_time=None, proof=None):
        self.root_hash = root_hash
        self.value = value
        self.seq_no = seq_no
        self.update_time = update_time
        self.proof = proof


def validate_attrib_keys(operation):
    data_keys = {RAW, ENC, HASH}.intersection(set(operation.keys()))
    return len(data_keys) == 1
